import abc as _abc
import copy as _cp
import pathlib as _pl
import re as _re

import pandas as _pd

import optihood.energy_network as en
import optihood.entities as _ent
from optihood.IO import readers as re
from optihood.Visualizer import convert_scenario as _cs, visualizer_app as _va
from optihood.energy_network import OptimizationProperties, EnergyNetworkClass as EnergyNetwork


# TODO: Figure out if the visualizer should stay here, or be called from the Network class
#       The User would not have access to the nodal_data at this time.

# TODO: EnergyNetworkIndiv vs EnergyNetworkGroup


class MpcComponentBasic:
    sheet_name: str
    main_labels: list[str]
    default_values: dict

    def maybe_get_entries_or_defaults(self, nodal_data: dict[str, _pd.DataFrame]) -> dict:
        df = nodal_data[self.sheet_name]
        obtained_rows = df[_ent.CommonLabels.label].isin(self.main_labels)
        if not any(obtained_rows):
            return {}

        initial_state = {}
        if self.required_entries_not_in_data(df):
            [initial_state.update({row[_ent.CommonLabels.label_unique]: self.default_values}) for i, row in
             df[obtained_rows].iterrows()]

            return initial_state

        for i, row in df[obtained_rows].iterrows():
            initial_state = self.get_entries(initial_state, row)

        return initial_state

    @staticmethod
    @_abc.abstractmethod
    def required_entries_not_in_data(df) -> bool:  # pragma: no cover
        """As this is always MPC, we will be overwriting these values at all time-steps.
        Thus, the User should always provide all required values.
        When the User misses one, or all, the defaults should be provided instead.
        """
        raise NotImplementedError

    @staticmethod
    @_abc.abstractmethod
    def get_entries(initial_state: dict, row) -> dict:  # pragma: no cover
        raise NotImplementedError


class StoragesMPC(MpcComponentBasic):
    main_labels = [_ent.StorageTypes.shStorage, _ent.StorageTypes.dhwStorage, _ent.StorageTypes.electricalStorage,
                   _ent.StorageTypes.coolingBufferStorage]
    sheet_name = _ent.NodeKeys.storages
    default_values = {_ent.StorageLabels.initial_capacity.value: 0.0}

    @staticmethod
    def required_entries_not_in_data(df) -> bool:
        return _ent.StorageLabels.initial_capacity not in df.columns

    @staticmethod
    def get_entries(initial_state: dict, row) -> dict:
        initial_state[row[_ent.CommonLabels.label_unique]] = {
            _ent.StorageLabels.initial_capacity.value: row[_ent.StorageLabels.initial_capacity]}
        return initial_state


class IceStorageMPC(MpcComponentBasic):
    sheet_name = _ent.NodeKeys.storages
    main_labels = [_ent.IceStorageTypes.iceStorage]
    default_values = {_ent.StorageLabels.initial_capacity.value: 0.0, _ent.StorageLabels.initial_temp.value: 2.0}

    @staticmethod
    def required_entries_not_in_data(df) -> bool:
        """As this is always MPC, we will be overwriting these values at all time-steps.
        Thus, the User should always provide all required values.
        When the User misses one, or all, the defaults should be provided instead.
        """
        return _ent.StorageLabels.initial_capacity not in df.columns and _ent.StorageLabels.initial_temp not in df.columns

    @staticmethod
    def get_entries(initial_state: dict, row) -> dict:
        initial_state[row[_ent.CommonLabels.label_unique]] = {
            _ent.StorageLabels.initial_capacity.value: row[_ent.StorageLabels.initial_capacity],
            _ent.StorageLabels.initial_temp.value: row[_ent.StorageLabels.initial_temp],
        }
        return initial_state


class BuildingMPC(MpcComponentBasic):
    """These need to be defined in a separate file.
    This file is declared in the profiles sheet
    As part of MPC, this file should already have been read in before running this.
    """
    sheet_name = _ent.NodeKeysOptional.building_model_parameters
    main_labels = [_ent.IceStorageTypes.iceStorage]
    default_values = {_ent.BuildingModelParameters.tIndoorInit.value: 20.0,
                      _ent.BuildingModelParameters.tDistributionInit.value: 15.0,
                      _ent.BuildingModelParameters.tWallInit.value: 25.0,
                      }

    def maybe_get_entries_or_defaults(self, nodal_data: dict[str, _pd.DataFrame]) -> dict:
        df = nodal_data[self.sheet_name]
        if df.empty:
            raise ValueError("Building model parameters should not be empty. Either add buildings, or do not include "
                             "the parameters to get_mpc_iputs")

        initial_state = {}
        if self.required_entries_not_in_data(df):
            [initial_state.update({row[_ent.BuildingModelParameters.building_unique.value]: self.default_values}) for
             i, row in df.iterrows()]

            return initial_state

        for i, row in df.iterrows():
            initial_state = self.get_entries(initial_state, row)

        return initial_state

    @staticmethod
    def required_entries_not_in_data(df) -> bool:
        """As this is always MPC, we will be overwriting these values at all time-steps.
        Thus, the User should always provide all required values.
        When the User misses one, or all, the defaults should be provided instead.
        """
        return (_ent.BuildingModelParameters.tIndoorInit not in df.columns
                and _ent.BuildingModelParameters.tDistributionInit not in df.columns
                and _ent.BuildingModelParameters.tWallInit not in df.columns
                )

    @staticmethod
    def get_entries(initial_state: dict, row) -> dict:
        initial_state[row[_ent.BuildingModelParameters.building_unique]] = {
            _ent.BuildingModelParameters.tIndoorInit.value: row[_ent.BuildingModelParameters.tIndoorInit],
            _ent.BuildingModelParameters.tDistributionInit.value: row[_ent.BuildingModelParameters.tDistributionInit],
            _ent.BuildingModelParameters.tWallInit.value: row[_ent.BuildingModelParameters.tWallInit],
        }
        return initial_state


def get_MPC_components_minimal():
    MPC_COMPONENTS: list[type[MpcComponentBasic]] = [
        StoragesMPC,
        IceStorageMPC,
    ]
    return MPC_COMPONENTS


def prep_mpc_inputs(nodal_data: dict[str, _pd.DataFrame],
                    building_model_parameters: _pd.DataFrame | None = None,
                    get_mpc_components=get_MPC_components_minimal
                    ) -> tuple[dict, dict]:
    MPC_COMPONENTS = get_mpc_components()
    initial_state_with_all_configurable_options = {}
    label_to_sheet = {}
    if building_model_parameters is not None:
        MPC_COMPONENTS.append(BuildingMPC)
        nodal_data[_ent.NodeKeysOptional.building_model_parameters] = building_model_parameters

    for i, component in enumerate(MPC_COMPONENTS):
        initial_states_for_component = component().maybe_get_entries_or_defaults(nodal_data)
        initial_state_with_all_configurable_options.update(initial_states_for_component)
        label_to_sheet_for_component = build_label_to_sheet(initial_states_for_component, component.sheet_name)
        label_to_sheet.update(label_to_sheet_for_component)

    return initial_state_with_all_configurable_options, label_to_sheet


def build_label_to_sheet(initial_states_for_component: dict[str: float], sheet_name: str):
    label_to_sheet_for_component = {}
    [label_to_sheet_for_component.update({label: sheet_name}) for label in initial_states_for_component.keys()]
    return label_to_sheet_for_component


class MpcHandler:
    """
    Class to simplify interaction with MPC functionality.
    It has the following responsibilities:
    - preparing the scenario from file.
    - providing the scenario's initial system state for the User to adjust.
    - updating the network with the current system state.
    - adjusting the prediction period to the current time-step.
    """
    time_period_full: _pd.DatetimeIndex
    nodal_data: dict  # change to class with dfs.
    optimization_settings: OptimizationProperties  # provide string literals?
    label_to_sheet: dict[str, str]

    def __init__(self, prediction_window_in_hours: int, time_step_in_minutes: int, nr_of_buildings: int,) -> None:
        self.nr_of_buildings = nr_of_buildings
        self.time_step_in_minutes = time_step_in_minutes
        self.prediction_window_in_hours = prediction_window_in_hours

    def visualize_example(self, node_layout_file: str = None) -> None:
        converters = _cs.get_converters(self.nodal_data, nr_of_buildings=self.nr_of_buildings)
        graphData = _cs.get_graph_data(converters)
        _va.run_cytoscape_visualizer(graphData=graphData, node_layout_file=node_layout_file)

    def update_network(self, current_time_step, current_state):
        return self._get_network(current_time_step, current_state)

    def _get_network(self, current_time_step, current_state):
        """Temporary function, which slowly recreates the same network over and over.
        This will need to be replaced after rapid refactoring.
        Then, the minor changes required will be applied into the relevant network components directly.
        """
        current_nodal_data = self.update_nodal_data(current_state)
        current_time_period = self.get_current_time_period(current_time_step)
        # TODO: allow user to adjust this, or ensure only one exists.
        network = EnergyNetwork(current_time_period)

        network.check_mutually_exclusive_inputs(self.optimization_settings.merge_link_buses)
        dispatch_mode = self.optimization_settings.dispatch_mode
        if not dispatch_mode:
            raise ValueError("dispatch_mode has to be true for MPC.")

        network._dispatchMode = dispatch_mode
        network._optimizationType = self.optimization_settings.optimization_type
        network._mergeBuses = self.optimization_settings.merge_buses

        network.set_using_nodal_data(
            nodesData=current_nodal_data,
            clusterSize=self.optimization_settings.cluster_size,
            filePath="",  # This is only used in a logging message after processing the data.
            includeCarbonBenefits=self.optimization_settings.include_carbon_benefits,
            mergeBuses=self.optimization_settings.merge_buses,
            mergeHeatSourceSink=self.optimization_settings.merge_heat_source_sink,
            mergeLinkBuses=self.optimization_settings.merge_link_buses,
            numberOfBuildings=self.nr_of_buildings,
            opt=self.optimization_settings.optimization_type,
        )
        return network

    def get_mpc_scenario_from_csv(self, input_folder_path: _pl.Path) -> dict[str, dict[str, float]]:
        csvReader = re.CsvScenarioReader(input_folder_path)
        nodal_data = csvReader.read_scenario()
        nodal_data = re.add_unique_label_columns(nodal_data)
        nodal_data = re.ProfileAndOtherDataReader().read_profiles_and_other_data(
            nodal_data, input_folder_path, self.nr_of_buildings, self.optimization_settings.cluster_size,
            self.time_period_full,
        )
        # TODO: adjust building_model_parameters entry to check nodal_data instead.
        system_state, label_to_sheet = prep_mpc_inputs(nodal_data, building_model_parameters=None)
        self.nodal_data = nodal_data
        self.label_to_sheet = label_to_sheet

        return system_state

    def update_nodal_data(self, current_state: dict[str, dict[str, float]]) -> dict:
        """Requires self.nodal_data"""
        # TODO: clip profiles?
        # TODO: adjust building model parameters to correct implementation.
        nodal_data = _cp.deepcopy(self.nodal_data)
        for label, inputs in current_state.items():
            sheet_name = self.label_to_sheet[label]
            sheet = nodal_data[sheet_name]
            label_column = _ent.CommonLabels.label_unique
            if sheet_name == _ent.NodeKeysOptional.building_model_parameters:
                label_column = _ent.BuildingModelParameters.building_unique
            row_index = sheet[label_column] == label
            for column_name, value in inputs.items():
                # TODO: deal with incompatible dtype. Sometimes int instead of float.
                sheet.loc[row_index, column_name] = value

        return nodal_data

    def get_desired_energy_flows(self, results: dict[str: _pd.DataFrame], desired_flows_with_new_names: dict[str, str]) -> _pd.DataFrame:
        results = self.rename_result_labels(results)
        flow_label_to_sheet = self.get_flow_label_to_sheet(results)
        index = self.get_date_time_index(results)
        energy_flows = _pd.DataFrame(index=index)
        for flow_label, new_label in desired_flows_with_new_names.items():
            sheet = flow_label_to_sheet[flow_label]
            energy_flows[new_label] = results[sheet][flow_label]

        return energy_flows

    def rename_result_labels(self, results: dict[str: _pd.DataFrame]) -> dict[str: _pd.DataFrame]:
        """
        Renames: (('electricityInBus__Building1', 'HP__Building1'), 'flow')
        To: electricityInBus__B001__To__HP__B001
        """
        for sheet, df in results.items():
            df.columns = [str(column) for column in df.columns]
            rename_dict = self.rename_oemof_labels(df.columns)
            df.rename(columns=rename_dict, inplace=True)

        return results

    @staticmethod
    def rename_oemof_labels(columns: list[str]) -> dict[str, str]:
        """
        Renames: (('electricityInBus__Building1', 'HP__Building1'), 'flow')
        To: electricityInBus__B001__To__HP__B001

        Ignores anything without 'flow', e.g.
        - "storage_content"
        - "some_flow_to_be_ignored"

        """
        # TODO: find a better place for this, as it should be used in the new results data class.
        rename_dict = {}
        for column in columns:
            column = str(column)
            if isinstance(column, int) or "'flow'" not in column:
                """Leave the column out, and it will stay the same."""
                continue
            parts = column.split(',')
            from_node = MpcHandler.rename_node(parts[0])
            to_node = MpcHandler.rename_node(parts[1])

            new_name = f"{from_node}__To__{to_node}"
            rename_dict[column] = new_name

        return rename_dict

    @staticmethod
    def rename_node(part: str) -> str:
        node_name = (
            part.replace("((", "")
            .replace(")", "")
            .replace("'", "")
            .replace(" ", "")
        )
        match = _re.search(r"Building(\d+)", node_name)
        if match:
            building_nr = match.group(1)
            node_name = node_name.replace(f"uilding{building_nr}", f'{str(building_nr).zfill(3)}')
        return node_name

    @staticmethod
    def get_flow_label_to_sheet(results):
        flow_label_to_sheet = {}
        for sheet, df in results.items():
            current_labels = df.columns
            [flow_label_to_sheet.update({label: sheet}) for label in current_labels]
        return flow_label_to_sheet

    @staticmethod
    def get_date_time_index(results: dict[str: _pd.DataFrame]) -> _pd.DatetimeIndex:
        index = None
        for sheet, df in results.items():
            match = _re.search("Bus__", sheet)
            if match:
                index = df.index
                break

        if isinstance(index, _pd.DatetimeIndex):
            return index

        raise ValueError("No DateTimeIndex found in results.")

    def set_full_time_period(self, start_year: int, start_month: int, start_day: int, end_year: int, end_month: int,
                             end_day: int) -> None:
        """Prepares date time indices starting at 00:00:00 on the start day and ending at 23:00:00 on the end day."""
        self.time_period_full = _pd.date_range(f"{start_year}-{start_month}-{start_day} 00:00:00",
                                               f"{end_year}-{end_month}-{end_day} 23:00:00",
                                               freq=f"{str(self.time_step_in_minutes)}min")

    def get_current_time_period(self, current_time_period_start: _pd.DatetimeIndex) -> _pd.DatetimeIndex:
        """Creates the time period for the current prediction window."""
        current_time_period_end = current_time_period_start + _pd.Timedelta(hours=self.prediction_window_in_hours)
        current_time_period = _pd.date_range(current_time_period_start, current_time_period_end,
                                             freq=f"{str(self.time_step_in_minutes)}min")
        return current_time_period

    @staticmethod
    def log_processing(network: en.EnergyNetworkClass, costs: bool = False, env_impacts: bool = False,
                       meta: bool = False) -> None:
        # TODO: redo when new results output is ready.
        if costs:
            network.log_costs()
        if env_impacts:
            network.log_environmental_impacts()
        if meta:
            network.log_meta_results()



import logging
from optihood._helpers import *
from oemof.thermal.stratified_thermal_storage import (
    calculate_storage_u_value,
    calculate_losses,
)
from oemof.network import network as on
from pyomo.core.base.block import ScalarBlock
from pyomo.environ import BuildAction
from pyomo.environ import Constraint
from pyomo.environ import Expression
from pyomo.environ import NonNegativeReals, Reals, Boolean, Binary
from pyomo.environ import Set
from pyomo.environ import Var
import oemof.solph as solph
from oemof.solph._plumbing import sequence as solph_sequence
from optihood.links import LinkStorageDummyInput, Link
from optihood.genericstorage_pit import GenericStoragePit
import numpy as np

class ElectricalStorage(solph.components.GenericStorage):
    def __init__(self, buildingLabel, input, output, loss_rate, initial_storage, efficiency_in, efficiency_out,
                 capacity_min, capacity_max, epc, base, varc, env_flow, env_capa, dispatchMode):
        if dispatchMode:
            investArgs = {'minimum':capacity_min,
                    'maximum':capacity_max,
                    'ep_costs':epc,
                    'custom_attributes': {'env_per_capa': env_capa}}
        else:
            investArgs = {'minimum':capacity_min,
                    'maximum':capacity_max,
                    'ep_costs':epc,
                    'existing':0,
                    'nonconvex':True,
                    'offset':base,
                    'custom_attributes': {'env_per_capa': env_capa}}
        super(ElectricalStorage, self).__init__(
            label="electricalStorage"+'__'+buildingLabel,
            inputs={
                input: solph.Flow(investment=solph.Investment(ep_costs=0)),
            },
            outputs={
                output: solph.Flow(investment=solph.Investment(ep_costs=0), variable_costs=varc, custom_attributes=
                {'env_per_flow':env_flow} )
            },
            loss_rate=loss_rate,
            initial_storage_level=initial_storage,
            inflow_conversion_factor=efficiency_in,
            outflow_conversion_factor=efficiency_out,
            balanced=False,
            invest_relation_input_capacity=efficiency_in,
            invest_relation_output_capacity=efficiency_out,
            investment=solph.Investment(**investArgs),
        )

class ThermalStorage(solph.components.GenericStorage):
    def __init__(self, label_str, stratifiedStorageParams, input, output, initial_storage, min, max, volume_cost, base,
                 varc, env_flow, env_cap, dispatchMode, is_tank = True, rho= 1, c=4.186):
        label = LabelStringManipulator(label_str)
        loss_rate, fixed_losses_relative, fixed_losses_absolute, capacity_min, capacity_max, epc, env_capa = \
            self._precalculate(stratifiedStorageParams,label,min,max,volume_cost,env_cap, is_tank=
            is_tank, rho=rho, c=c)
        storageLabel = label.strip_trailing_digits_from_prefix()
        if dispatchMode:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'custom_attributes': {'env_per_capa': env_capa}}
        else:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'existing':0,
                'nonconvex':True,
                'offset':base,
                'custom_attributes': {'env_per_capa': env_capa}}

        super(ThermalStorage, self).__init__(
            label=label.full_name,
            inputs={
                input: solph.Flow(investment=solph.Investment(ep_costs=0)),
            },
            outputs={
                output: solph.Flow(investment=solph.Investment(ep_costs=0), variable_costs=varc, custom_attributes=
                {'env_per_flow':env_flow} )
            },
            loss_rate=loss_rate,
            initial_storage_level=initial_storage,
            fixed_losses_relative=fixed_losses_relative,
            fixed_losses_absolute=fixed_losses_absolute,
            inflow_conversion_factor=stratifiedStorageParams.at[storageLabel, 'inflow_conversion_factor'],
            outflow_conversion_factor=stratifiedStorageParams.at[storageLabel, 'outflow_conversion_factor'],
            invest_relation_input_capacity=1,
            invest_relation_output_capacity=1,
            balanced=False,
            investment=solph.Investment(**investArgs),
        )

    def _precalculate(self, data, label,min,max,volume_cost,env_cap, is_tank, rho, c):
        label_prefix = label.prefix
        if is_tank:
            tempH = data.at[label_prefix, 'temp_h']
            tempC = data.at[label_prefix, 'temp_c']
            u_value = calculate_storage_u_value(
                data.at[label_prefix, 's_iso'],
                data.at[label_prefix, 'lamb_iso'],
                data.at[label_prefix, 'alpha_inside'],
                data.at[label_prefix, 'alpha_outside'])

            loss_rate, fixed_losses_relative, fixed_losses_absolute = calculate_losses(
                u_value,
                data.at[label_prefix, 'diameter'],
                tempH,
                tempC,
                data.at[label_prefix, 'temp_env'])

            L_to_kWh = c * rho * (tempH - tempC) / 3600  # converts L data to kWh data for oemof GenericStorage class
        else:
            time_increment =1
            # TODO: refactor into functions
            if 'GenericStorage' in label_prefix:
                typical_losses = {'tank': 0.10, 'pit': 0.20, 'borehole': 0.15, 'aquifier': 0.001}
                try:
                    loss_rate = next(value for key, value in typical_losses.items() if key in label_prefix)
                except StopIteration:
                    raise ValueError(
                        f"No matching storage type found in label '{label_prefix}'. The label should contain one of: {list(typical_losses.keys())}")
                fixed_losses_relative = 0
                fixed_losses_absolute = 0
                energy_density = data.at[label.strip_trailing_digits_from_prefix(), 'energy_density_per_m3']
                L_to_kWh = energy_density/1000 # m3 to L conversion
            else:
                temp_h = data.at[label_prefix, 'temp_h']
                temp_c = data.at[label_prefix, 'temp_c']
                temp_env = data.at[label_prefix, 'temp_env']
                rho = data.at[label_prefix, 'rho']
                c = data.at[label_prefix, 'c']
                u_value = data.at[label_prefix, 'u_value']

                if 'pitStorage' in label_prefix:
                    height = data.at[label_prefix, 'height']
                    angle = data.at[label_prefix, 'angle']
                    fixed_losses_relative = 3 * u_value * (temp_c - temp_env) / (height * c * rho * (temp_h - temp_c)) * time_increment *3600
                    loss_rate = 0  # 0.4 * fixed_losses_relative#2 * u_value / (s_base * c * rho * math.sin(angle * np.pi / 180)) * time_increment
                    fixed_losses_absolute = 0
                else:
                    loss_rate = 0
                    fixed_losses_relative = 0
                    fixed_losses_absolute = u_value*(temp_h - temp_env)*1e-6 #convert Wh to MWh
                L_to_kWh = c * rho * (temp_h - temp_env) / 3600  # converts L data to kWh data for oemof GenericStorage class

        capacity_min = min * L_to_kWh
        capacity_max = max * L_to_kWh
        epc = volume_cost / L_to_kWh
        env_capa = env_cap / L_to_kWh

        return loss_rate, fixed_losses_relative, fixed_losses_absolute, capacity_min, capacity_max, epc, env_capa

class ThermalStoragePit(GenericStoragePit):
    def __init__(self, label, stratifiedStorageParams, input, output, initial_storage, min, max, volume_cost, base,
                 varc, env_flow, env_cap, dispatchMode, rho= 1, c=4.186):

        height_multiplication_factor, loss_rate, fixed_losses_relative, capacity_min, capacity_max, epc, env_capa, \
        temp_h, temp_c = \
            self._precalculate(stratifiedStorageParams,label.split("__")[0].rstrip('0123456789'),min,max,volume_cost,env_cap, rho=rho, c=c)

        storageLabel = label.split("__")[0]

        if dispatchMode:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'custom_attributes': {'env_per_capa': env_capa}}
        else:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'existing':0,
                'nonconvex':True,
                'offset':base,
                'custom_attributes': {'env_per_capa': env_capa}}

        super(ThermalStoragePit, self).__init__(
            label=label,
            inputs={
                input: solph.Flow(investment=solph.Investment(ep_costs=0)),
            },
            outputs={
                output: solph.Flow(investment=solph.Investment(ep_costs=0), variable_costs=varc, custom_attributes=
                {'env_per_flow':env_flow} )
            },
            loss_rate=loss_rate,
            initial_storage_level=initial_storage,
            fixed_losses_relative=fixed_losses_relative,
            fixed_losses_absolute=0,
            height_multiplication_factor=height_multiplication_factor,
            temp_h=temp_h,
            temp_c=temp_c,
            rho=rho,
            c=c,
            inflow_conversion_factor=stratifiedStorageParams.at[storageLabel.rstrip('0123456789'), 'inflow_conversion_factor'],
            outflow_conversion_factor=stratifiedStorageParams.at[storageLabel.rstrip('0123456789'), 'outflow_conversion_factor'],
            invest_relation_input_capacity=1,
            invest_relation_output_capacity=1,
            balanced=False,
            investment=solph.Investment(**investArgs),
        )

    def _precalculate(self, data, label, min, max, volume_cost, env_cap, rho, c, time_increment=1):
        temp_h = data.at[label, 'temp_h']
        temp_c = data.at[label, 'temp_c']
        temp_env = data.at[label, 'temp_env']
        temp_ground = data.at[label, 'temp_ground']
        alpha_inside = data.at[label, 'alpha_inside']
        alpha_outside = data.at[label, 'alpha_outside']
        s_iso_top = data.at[label, 's_iso']
        lamb_iso_top = data.at[label, 'lamb_iso']
        s_iso_bot = data.at[label, 's_iso_bot']
        lamb_iso_bot = data.at[label, 'lamb_iso_bot']

        # U-value for the top insulation
        denominator_top = 1 / alpha_inside + s_iso_top * 1e-3 / lamb_iso_top + 1 / alpha_outside
        u_value_top = 1 / denominator_top  # W/(m2*K)

        # U-value for the wall/bottom insulation
        denominator_bot = 1 / alpha_inside + s_iso_bot * 1e-3 / lamb_iso_bot + 1 / alpha_outside
        u_value_wall_bot = 1 / denominator_bot  # W/(m2*K)

        # Height multiplication factor
        height_multiplication_factor = (151 / 3) * rho * c * 1e6 * (temp_h - temp_c)

        # Loss rate calculation
        loss_rate = (u_value_wall_bot * 84 * np.sqrt(5) * time_increment * 3600) / (151 * rho * c * 1e6)  # m

        # Fixed losses relative calculation
        fixed_losses_relative = (u_value_top * 243 * (temp_h - temp_env) * time_increment * 3600
                                 + u_value_wall_bot * 75 * (temp_c - temp_ground) * time_increment * 3600
                                 + u_value_wall_bot * 84 * np.sqrt(5) * (temp_c - temp_ground)*time_increment*3600)\
                                 / (151 * rho * c * 1e6 * (temp_h - temp_c))  # m

        L_to_kWh = c * rho * (temp_h - temp_c) / 3600  # converts L data to kWh data for oemof GenericStorage class
        capacity_min = min * L_to_kWh
        capacity_max = max * L_to_kWh
        epc = volume_cost / L_to_kWh
        env_capa = env_cap / L_to_kWh

        return height_multiplication_factor, loss_rate, fixed_losses_relative, capacity_min, capacity_max, epc, \
            env_capa, temp_h, temp_c

class ThermalStorageTemperatureLevels:
    def __init__(self, label, stratifiedStorageParams, inputs, outputs, initial_storage, min, max, volume_cost, base, varc, env_flow, env_cap, dispatchMode):
        self._label = label
        storageLabel = label.split("__")[0]
        self.__TempH, self.__TempC = self._calcTemperatures(stratifiedStorageParams, storageLabel)
        capacity_min, capacity_max, epc, env_capa = self._conversionCapacities(self.__TempH[0], self.__TempC[0], min, max, volume_cost, env_cap)
        # capacity_min, capacity_max should be on the whole storage not individual temperature levels
        self.__capacityMin = capacity_min
        self.__capacityMax = capacity_max
        #self.__baseInvestment=base # offset should be added only once for the whole storage, not for each temperature level
        self._numberOfLevels = len(stratifiedStorageParams.at[storageLabel, 'temp_h'].split(","))
        if dispatchMode:
            investArgs={'ep_costs':epc,
                'custom_attributes': {'env_per_capa': env_capa},
                'minimum': self.__capacityMin / self._numberOfLevels,
                'maximum': self.__capacityMax}
        else:
            investArgs={'ep_costs':epc,
                'existing':0,
                'minimum': self.__capacityMin/self._numberOfLevels,
                'maximum':self.__capacityMax,
                'offset':0,
                'custom_attributes': {'env_per_capa': env_capa}}
        self._s = []  # list of storage volumes
        self._dummyComponents = [] # list of dummy inputs for flow of volume from storage level at lower temperature to higher temperature
        self._dummyInBus = []
        self._dummyOutBus = []
        for i in range(self._numberOfLevels):
            u_value, loss_rate, fixed_losses_relative, fixed_losses_absolute = self._precalculate(stratifiedStorageParams, label.split("__")[0], i)
            newLabel = storageLabel + str(int(self.__TempH[i])) + "__" + label.split("__")[1]
            if i == 0:
                storageInput = {inputs[i]: solph.Flow(investment=solph.Investment(ep_costs=0))}
                dummyInBus = solph.Bus(label="dummyInBus_" + newLabel.replace("thermalStorage", "ts"))
                dummyOutBus = solph.Bus(label="dummyOutBus_" + newLabel.replace("thermalStorage", "ts"))   #Special bus to dummy component for level 0 as GenericStorage can have only one output
                self._dummyInBus.append(dummyInBus)
                self._dummyOutBus.append(dummyOutBus)
                storageOutput = {dummyOutBus: solph.Flow(investment=solph.Investment(ep_costs=0))}
                specialOut = {outputs[0]: solph.Flow(),
                              dummyInBus: solph.Flow()}
                self._dummyComponents.append(Link(                              # special component at the output of storage level 0
                    label="dummy_"+newLabel,
                    inputs={dummyOutBus: solph.Flow()},
                    outputs=specialOut,
                    conversion_factors={b: 1 for b in specialOut}
                ))
            else:
                if i != self._numberOfLevels - 1:
                    dummyInBus = solph.Bus(label="dummyInBus_"+newLabel.replace("thermalStorage", "ts"))
                    self._dummyInBus.append(dummyInBus)
                dummyOutBus = solph.Bus(label="dummyOutBus_" + newLabel.replace("thermalStorage", "ts"))
                self._dummyOutBus.append(dummyOutBus)
                storageInput = {dummyOutBus: solph.Flow(investment=solph.Investment(ep_costs=0))}
                if i == self._numberOfLevels - 1:
                    storageOutput = {
                        outputs[-1]: solph.Flow(investment=solph.Investment(ep_costs=0), variable_costs=varc,
                                               custom_attributes={'env_per_flow':env_flow})}
                    investArgs.update({"offset": base, "nonconvex": True})
                else:
                    storageOutput = {dummyInBus: solph.Flow(investment=solph.Investment(ep_costs=0))}
                self._dummyComponents.append(LinkStorageDummyInput(label="dummy_"+newLabel,
                            inputs={inputs[i]: solph.Flow(),
                                    self._dummyInBus[i-1]: solph.Flow()},
                            outputs={dummyOutBus: solph.Flow()},
                            conversion_factors={k: 1 for k in [inputs[i], self._dummyInBus[i-1]]}))
            self._s.append(solph.components.GenericStorage(label=newLabel,
                                                inputs=storageInput,
                                                outputs=storageOutput,
                                                loss_rate=loss_rate,
                                                initial_storage_level=initial_storage,      # initial storage is applied to each temperature level (if specified)
                                                fixed_losses_relative=fixed_losses_relative,
                                                fixed_losses_absolute=fixed_losses_absolute,
                                                inflow_conversion_factor=stratifiedStorageParams.at[storageLabel, 'inflow_conversion_factor'],
                                                outflow_conversion_factor=stratifiedStorageParams.at[storageLabel, 'outflow_conversion_factor'],
                                                invest_relation_input_capacity=1,
                                                invest_relation_output_capacity=1,
                                                balanced=False,
                                                investment=solph.Investment(**investArgs),))

    @property
    def label(self):
        return self._label

    @property
    def capacityMin(self):
        return self.__capacityMin

    @property
    def capacityMax(self):
        return self.__capacityMax

    """@property
    def baseInvestment(self):
        return self.__baseInvestment"""

    def _calcTemperatures(self, data, label):
        tempH = [float(t) for t in data.at[label, 'temp_h'].split(",")]
        tempC = [float(data.at[label, 'temp_c'])]
        for i in range(len(tempH)):
            if i != len(tempH)-1:
                tempC.append(tempH[i])
        return tempH, tempC

    def _conversionCapacities(self, tempH, tempC, min, max, volume_cost, env_cap):
        L_to_kWh = 4.186 * (tempH - tempC) / 3600  # converts L data to kWh data for oemof GenericStorage class
        capacity_min = min * L_to_kWh
        capacity_max = max * L_to_kWh
        epc = volume_cost / L_to_kWh
        env_capa = env_cap / L_to_kWh
        return capacity_min, capacity_max, epc, env_capa

    def _precalculate(self, data, label, level):
        u_value = calculate_storage_u_value(
            data.at[label, 's_iso'],
            data.at[label, 'lamb_iso'],
            data.at[label, 'alpha_inside'],
            data.at[label, 'alpha_outside'])
        loss_rate, fixed_losses_relative, fixed_losses_absolute = calculate_losses(
            u_value,
            data.at[label, 'diameter'],
            self.__TempH[level],
            self.__TempC[level],
            data.at[label, 'temp_env'])
        if level==0 or level==self._numberOfLevels-1:
            fixed_losses_absolute /= 2
        else:
            fixed_losses_absolute = 0
        return u_value, loss_rate, fixed_losses_relative, fixed_losses_absolute

    def getStorageLevel(self, level):
        if level>=self._numberOfLevels:
            logging.error("Inconsistency in number of storage level buses and temp_h values.")
        else:
            return self._s[level]

    def getDummyComponents(self, level):
        if level>=self._numberOfLevels:
            logging.error("Inconsistency in number of storage level buses and temp_h values.")
        elif level == self._numberOfLevels - 1:
            return [self._dummyOutBus[level], self._dummyComponents[level]]
        else:
            return [self._dummyInBus[level], self._dummyOutBus[level], self._dummyComponents[level]]

class IceStorage(on.Node):
    r"""
    Models the basic characteristics of an ice storage component
    IceStorage is designed with one input and one output

    Parameters
    ---------------------------------------------------------
    label: label of the ice storage node
    input: input bus of the ice storage node
    output: output bus of the ice storage node
    tStorInit: initial temperature of the ice storage [°C]
    fMax: maximum allowed ice fraction in the ice storage
    rho: density of water [kg/m3]
    V: Volume of the ice storage [m3]
    hfluid: latent heat of fusion of water [kJ/kg]
    cp: specific heat capacity of water [kJ/(kg.°C)]
    T_amb: ambient temperature time-series [°C]
    UA_tank: heat loss coefficient of the ice storage tank [kW/°C]
    inflow_conversion_factor: efficiency of heat exchanger at the inlet of the ice storage
    outflow_conversion_factor: efficiency of heat exchanger at the outlet of the ice storage
    """
    def __init__(self, label, input, output, tStorInit, fIceInit, fMax, rho, V, hfluid, cp, Tamb, UAtank, inflow_conversion_factor, outflow_conversion_factor):
        if tStorInit < 0:
            raise ValueError("The initial temperature of ice storage should be greater than or equal to 0°C.")
        self.tStorInit = tStorInit
        if fIceInit < 0 or fIceInit > 1:
            raise ValueError("The initial ice fraction in ice storage should be in [0,1]")
        self.fIceInit = fIceInit
        if fMax < 0.5 or fMax > 0.8:
            raise ValueError("Maximum ice fraction should be defined within the range 0.5-0.8")
        self.fMax = fMax
        self.massWaterMax = rho*V
        self.hf = hfluid
        self.rho = rho
        self.V = V
        self.cp = cp
        self.Tamb = solph_sequence(Tamb)
        self.UAtank = UAtank
        self.inflow_conversion_factor = inflow_conversion_factor
        self.outflow_conversion_factor = outflow_conversion_factor

        """if dispatchMode:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'custom_attributes': {'env_per_capa': env_capa}}
        else:
            investArgs={'minimum':capacity_min,
                'maximum':capacity_max,
                'ep_costs':epc,
                'existing':0,
                'nonconvex':True,
                'offset':base,
                'custom_attributes': {'env_per_capa': env_capa}}"""

        self.investment = None
        self._invest_group = False

        inputs = {input: solph.Flow(),}
        outputs = {output: solph.Flow()}

        super().__init__(
            label,
            inputs=inputs,
            outputs=outputs,
            custom_properties={},
        )

    def constraint_group(self):
        if self._invest_group is True:
            return IceStorageInvestmentBlock
        else:
            return IceStorageBlock

class IceStorageBlock(ScalarBlock):
    """
    Constraint block for ice storage without investment object
    """
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        if group is None:
            return None

        m = self.parent_block()

        i = {n: [i for i in n.inputs][0] for n in group}
        o = {n: [o for o in n.outputs][0] for n in group}

        #  ************* SET OF ICE STORAGES *****************************
        self.icestorages = Set(initialize=[n for n in group])

        #  ************* DECISION VARIABLES *****************************
        # temperature of ice storage
        self.tStor = Var(self.icestorages, m.TIMESTEPS, within=NonNegativeReals, bounds=(0,90))
        self.tStor_prev = Var(self.icestorages, m.TIMESTEPS, within=NonNegativeReals, bounds=(0,90))
        # mass of ice in storage
        self.mIceStor = Var(self.icestorages, m.TIMESTEPS, within=NonNegativeReals, bounds=(0,1000000))
        self.mIceStor_prev = Var(self.icestorages, m.TIMESTEPS, within=NonNegativeReals, bounds=(0,1000000))
        # ice fraction
        self.fIce = Var(self.icestorages, m.TIMESTEPS, within=NonNegativeReals, bounds=(0,1))
        # binary variable defining the status of ice formation in the storage
        self.iceStatus = Var(self.icestorages, m.TIMESTEPS, within=Binary)

        # for linearization of non-linear constraints with big M method
        M = 2000
        epsilon = 0.000001

        #  ************* CONSTRAINTS *****************************

        def _initial_temperature_rule(block):
            """set initial value of temperature of ice storage
            """
            for g in group:
                lhs = self.tStor_prev[g,0]
                rhs = g.tStorInit
                block.initial_temperature.add((g, 0), (lhs == rhs))

        self.initial_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.initial_temperature_build = BuildAction(rule=_initial_temperature_rule)

        def _initial_ice_state_rule(block):
            """set initial state of ice storage --> no ice formation
            """
            for g in group:
                lhs = self.mIceStor_prev[g,0]
                rhs = g.fIceInit*g.rho*g.V
                block.initial_ice_state.add((g, 0), (lhs == rhs))

        self.initial_ice_state = Constraint(group, m.TIMESTEPS, noruleinit= True)
        self.initial_ice_state_build = BuildAction(rule=_initial_ice_state_rule)

        def _prev_temperature_rule(block):
            for g in group:
                for t in m.TIMESTEPS:
                    if t != 0:
                        lhs = self.tStor_prev[g, t]
                        rhs = self.tStor[g, t - 1]
                        block.prev_temperature.add((g, t), (lhs == rhs))

        self.prev_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.prev_temperature_build = BuildAction(rule=_prev_temperature_rule)

        def _prev_mass_ice_rule(block):
            for g in group:
                for t in m.TIMESTEPS:
                    if t != 0:
                        lhs = self.mIceStor_prev[g, t]
                        rhs = self.mIceStor[g, t - 1]
                        block.prev_mass_ice.add((g, t), (lhs == rhs))

        self.prev_mass_ice = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.prev_mass_ice_build = BuildAction(rule=_prev_mass_ice_rule)

        def _max_ice_fraction_rule(block):
            """set the value of ice fraction"""
            for g in group:
                for t in m.TIMESTEPS:
                    lhs = self.fIce[g,t]
                    rhs = g.fMax
                    block.max_ice_fraction.add((g, t), (lhs <= rhs))

        self.max_ice_fraction = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.max_ice_fraction_build = BuildAction(rule=_max_ice_fraction_rule)

        def _ice_fraction_rule(block):
            """set the value of ice fraction"""
            for g in group:
                for t in m.TIMESTEPS:
                    lhs = self.fIce[g,t]
                    rhs = self.mIceStor[g,t]/g.massWaterMax     # division is most likely not allowed!!
                    block.ice_fraction.add((g, t), (lhs == rhs))

        self.ice_fraction = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.ice_fraction_build = BuildAction(rule=_ice_fraction_rule)

        def _storage_balance_rule(block):
            """rule defining the energy balance of an ice storage in every timestep"""
            for g in group:
                for t in m.TIMESTEPS:
                    expr = 0
                    expr += g.rho*g.V*g.cp*self.tStor[g,t]
                    expr += - g.rho * g.V * g.cp * self.tStor_prev[g, t]
                    expr += g.UAtank*self.tStor_prev[g,t]*m.timeincrement[t]
                    expr += - g.UAtank * g.Tamb[t] * m.timeincrement[t]
                    expr += - g.hf*self.mIceStor[g,t]
                    expr += g.hf * self.mIceStor_prev[g, t]
                    expr += (-m.flow[i[g], g, t]*g.inflow_conversion_factor*m.timeincrement[t])
                    expr += ((m.flow[g, o[g], t]/g.outflow_conversion_factor)*m.timeincrement[t])
                    block.storage_balance.add((g, t), (expr == 0))

        self.storage_balance = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.storage_balance_build = BuildAction(rule=_storage_balance_rule)

        def _mass_ice_rule(block):
            """rule for calculating the mass of ice in each timestep"""
            for g in group:
                for t in m.TIMESTEPS:
                    lhs = self.mIceStor[g, t]
                    rhs = self.iceStatus[g, t]*(self.mIceStor_prev[g, t] +
                                                (((m.flow[g, o[g], t]/g.outflow_conversion_factor)*m.timeincrement[t]
                                                  - m.flow[i[g], g, t]*g.inflow_conversion_factor*m.timeincrement[t]
                                                  + g.UAtank*(self.tStor_prev[g, t] - g.Tamb[t]) * m.timeincrement[t]
                                                  - g.rho*g.V*g.cp*self.tStor_prev[g, t])/g.hf))
                    block.mass_ice.add((g, t), (lhs == rhs))

        self.mass_ice = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.mass_ice_build = BuildAction(rule=_mass_ice_rule)

        def _ice_state_rule_1(block):
            """rule for calculating the mass of ice in each timestep"""
            for g in group:
                for t in m.TIMESTEPS:
                    lhs = self.tStor[g,t]
                    rhs = M*(1-self.iceStatus[g,t])
                    block.ice_state_1.add((g, t), (lhs <= rhs))

        self.ice_state_1 = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.ice_state_1_build = BuildAction(rule=_ice_state_rule_1)

        def _ice_state_rule_2(block):
            """rule for calculating the mass of ice in each timestep"""
            for g in group:
                for t in m.TIMESTEPS:
                    lhs = self.tStor[g,t]
                    rhs = epsilon*(1-self.iceStatus[g,t])
                    block.ice_state_2.add((g, t), (lhs >= rhs))

        self.ice_state_2 = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.ice_state_2_build = BuildAction(rule=_ice_state_rule_2)

        # def _max_heating_energy_rule(block):
        #     """rule for calculating the mass of ice in each timestep"""
        #     for g in group:
        #         for t in m.TIMESTEPS:
        #             lhs1 = m.flow[i[g], g, t]
        #             lhs2 = m.flow[g, o[g], t]
        #             rhs = 50
        #             block.max_heating_energy_in.add((g, t), (lhs1 <= rhs))
        #             block.max_heating_energy_out.add((g, t), (lhs2 <= rhs))
        #
        # self.max_heating_energy_in = Constraint(group, m.TIMESTEPS, noruleinit=True)
        # self.max_heating_energy_out = Constraint(group, m.TIMESTEPS, noruleinit=True)
        # self.max_heating_energy_build = BuildAction(rule=_max_heating_energy_rule)

    def _objective_expression(self):
        """objective expression for storages with no investment"""
        m = self.parent_block()
        if not hasattr(self, "icestorages"):
            return 0
        costs = 0   # no cost from using the storage with a fixed capacity
        self.costs = Expression(expr=costs)
        return self.costs

class IceStorageInvestmentBlock(ScalarBlock):
    """
    Constraint block for ice storage with investment attribute set as not None
    """


from pyomo.core.base.block import ScalarBlock
from pyomo.environ import BuildAction
from pyomo.environ import Constraint
from optihood._helpers import *
from oemof.solph import components as solph_components
from oemof.solph._plumbing import sequence as solph_sequence


class CombinedTransformer(solph_components.Transformer):
    r"""
    A transformer able to produce both SH and DHW in the same timestep
    Pelec_in = Qsh/efficiencySH + Qdhw/efficiencyDHW
    """

    def __init__(self, efficiencies, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.efficiency = {
            k: solph_sequence(v)
            for k, v in efficiencies.items()
        }

    def constraint_group(self):
        return CombinedTransformerBlock


class CombinedTransformerBlock(ScalarBlock):
    r"""Block for the linear relation of nodes
    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """Creates the linear constraint
        """
        if group is None:
            return None

        m = self.parent_block()

        for n in group:
            n.inflow = [i for i in list(n.inputs) if "electricity" in i.label or list(n.inputs).__len__()==1][0]
            if len(list(n.inputs)) > 1:
                n.inflowQevap = [i for i in list(n.inputs) if "electricity" not in i.label][0]
            else:
                n.inflowQevap = 0
            flows = [k for k, v in n.efficiency.items()]
            n.outputs_ordered = []
            n.efficiency_sq = ()
            for i in range(len(flows)):
                n.outputs_ordered.append([o for o in n.outputs if flows[i] == o][0])
                n.efficiency_sq = n.efficiency_sq + (n.efficiency[n.outputs_ordered[i]],)

        def _input_output_relation_rule(block):
            """Connection between input and outputs."""
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g.inflow, g, t]
                    rhs = sum(m.flow[g, g.outputs_ordered[i], t] / g.efficiency_sq[i][t] for i in range(len(g.efficiency_sq)))
                    block.input_output_relation.add((g, t), (lhs == rhs))

        self.input_output_relation = Constraint(
            group, m.TIMESTEPS, noruleinit=True
        )
        self.input_output_relation_build = BuildAction(
            rule=_input_output_relation_rule
        )

        def _second_input_relation_rule(block):
            """Constraint for evaluation of Q_condensor i.e. the second input"""
            for t in m.TIMESTEPS:
                for g in group:
                    if len(list(g.inputs)) > 1:
                        lhs = m.flow[g.inflowQevap, g, t]
                        rhs = (sum(m.flow[g, g.outputs_ordered[i], t] for i in range(len(g.efficiency_sq))) - m.flow[g.inflow, g, t])
                        block.input_relation.add((g, t), (lhs == rhs))

        self.input_relation = Constraint(
            group, m.TIMESTEPS, noruleinit=True
        )
        self.input_relation_build = BuildAction(
            rule=_second_input_relation_rule
        )


class CombinedCHP(solph_components.Transformer):
    r"""
    A CHP able to produce both SH and DHW in the same timestep
    Pelec_in = Qsh/efficiencySH + Qdhw/efficiencyDHW
             = Pelec_out/efficiencyEl
    """

    def __init__(self, efficiencies, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.efficiency = {
            k: solph_sequence(v)
            for k, v in efficiencies.items()
        }

    def constraint_group(self):
        return CombinedCHPBlock


class CombinedCHPBlock(ScalarBlock):
    r"""Block for the linear relation of nodes
    """

    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        """Creates the linear constraint
        """
        if group is None:
            return None

        m = self.parent_block()

        for n in group:
            n.inflow = list(n.inputs)[0]
            flows = [k for k, v in n.efficiency.items()]
            n.outputs_ordered = []
            n.efficiency_sq = ()
            for i in range(len(flows)):
                if "electricity" not in flows[i].label:
                    n.outputs_ordered.append([o for o in n.outputs if flows[i] == o][0])
                    n.efficiency_sq = n.efficiency_sq + (n.efficiency[n.outputs_ordered[-1]],)
            for i in range(len(flows)):
                if "electricity" in flows[i].label: # add electricity output flow and efficiency as the last element
                    n.outputs_ordered.append([o for o in n.outputs if flows[i] == o][0])
                    n.efficiency_sq = n.efficiency_sq + (n.efficiency[n.outputs_ordered[-1]],)
            print("")
        def _input_heat_relation_rule(block):
            """Connection between input and heat outputs."""
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g.inflow, g, t]
                    rhs = sum(m.flow[g, g.outputs_ordered[i], t] / g.efficiency_sq[i][t] for i in range(len(g.efficiency_sq)-1))
                    block.input_heat_relation.add((g, t), (lhs == rhs))

        self.input_heat_relation = Constraint(
            group, m.TIMESTEPS, noruleinit=True
        )
        self.input_heat_relation_build = BuildAction(
            rule=_input_heat_relation_rule
        )

        def _input_elec_relation_rule(block):
            """Connection between input and elec output."""
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g.inflow, g, t]
                    rhs = (m.flow[g, g.outputs_ordered[-1], t] / g.efficiency_sq[-1][t])
                    block.input_elec_relation.add((g, t), (lhs == rhs))

        self.input_elec_relation = Constraint(
            group, m.TIMESTEPS, noruleinit=True
        )
        self.input_elec_relation_build = BuildAction(
            rule=_input_elec_relation_rule
        )


import os as _os
import pathlib as _pl
import unittest as _ut

import pandas as _pd
import pytest as _pt

import optihood as _oh
import optihood.energy_network as _en
from tests.xls_helpers import check_assertion, check_dataframe_assertion

cwd = _os.getcwd()
packageDir = _pl.Path(_oh.__file__).resolve().parent
_input_data_dir = packageDir / ".." / "data" / "excels" / "basic_example"
_examples_dir = packageDir / ".." / "data" / "examples"
_input_data_path = _input_data_dir / "scenario.xls"
_input_csv_dir = packageDir / ".." / "data" / "CSVs" / "basic_example_CSVs"


class TestEnergyNetwork(_ut.TestCase):
    def setUp(self):
        self.maxDiff = None

    @_pt.mark.xdist_group('expensive_06')
    def test_set_from_excel(self):
        """File System touching test to check see whether network is populated correctly using a scenario."""

        # Given

        time_period = _pd.date_range("2018-01-01 00:00:00", "2018-01-31 23:00:00", freq="60min")
        nr_of_buildings = 4
        optimization_type = "costs"  # set as "env" for environmental optimization and "costs" for cost optimization

        # When
        _os.chdir(_examples_dir)
        network = _en.EnergyNetworkClass(time_period)
        network.setFromExcel(str(_input_data_path), nr_of_buildings, opt=optimization_type)
        _os.chdir(cwd)

        # Then
        errors = []
        check_assertion(self, errors, len(network.groups), 147)

        check_assertion(self, errors, len(network.nodes), 140)

        check_assertion(self, errors, network.results, None)

        check_assertion(self, errors, len(network.signals), 1)

        check_assertion(self, errors, network.temporal, None)

        check_assertion(self, errors, len(network.timeincrement), 744)

        check_assertion(self, errors, network.timeincrement.unique(), 1.0)

        check_assertion(self, errors, len(network.timeindex), 745)

        if errors:
            raise ExceptionGroup(f"found {len(errors)} errors", errors)

    def test_get_nodal_data_from_Excel(self):
        """File System touching test to get an initial grip on what the nodal data should be after reading the
        scenario."""
        network = _en.EnergyNetworkClass(None)

        data = _pd.ExcelFile(str(_input_data_path))
        initial_nodal_data = network.get_nodal_data_from_Excel(data)
        data.close()

        errors = []

        check_assertion(self, errors, len(initial_nodal_data), 10)
        check_assertion(
            self,
            errors,
            list(initial_nodal_data.keys()),
            [
                "buses",
                "grid_connection",
                "commodity_sources",
                "solar",
                "transformers",
                "demand",
                "storages",
                "stratified_storage",
                "profiles",
                "links"
            ],
        )

        check_assertion(self, errors, initial_nodal_data["buses"].shape, (48, 5))
        check_assertion(self, errors, initial_nodal_data["grid_connection"].shape, (24, 5))
        check_assertion(self, errors, initial_nodal_data["commodity_sources"].shape, (8, 6))
        check_assertion(self, errors, initial_nodal_data["solar"].shape, (8, 31))
        check_assertion(self, errors, initial_nodal_data["transformers"].shape, (16, 21))
        check_assertion(self, errors, initial_nodal_data["demand"].shape, (12, 7))
        check_assertion(self, errors, initial_nodal_data["storages"].shape, (12, 20))
        check_assertion(self, errors, initial_nodal_data["stratified_storage"].shape, (2, 11))
        check_assertion(self, errors, initial_nodal_data["profiles"].shape, (2, 3))
        check_assertion(self, errors, initial_nodal_data["links"].shape, (3, 6))

        if errors:
            raise ExceptionGroup(f"found {len(errors)} errors", errors)

    def test_set_from_csvs(self):
        """File system touching test.
        Compare network state of the csv variety with that of the Excel variety.
        """
        time_period = _pd.date_range("2018-01-01 00:00:00", "2018-01-31 23:00:00", freq="60min")
        nr_of_buildings = 4
        optimization_type = "costs"  # set as "env" for environmental optimization and "costs" for cost optimization

        # When
        _os.chdir(_examples_dir)
        network_csv = _en.EnergyNetworkClass(time_period)
        network_csv.set_from_csv(_input_csv_dir, nr_of_buildings, opt=optimization_type)

        network_excel = _en.EnergyNetworkClass(time_period)
        network_excel.setFromExcel(_input_data_path, nr_of_buildings, opt=optimization_type)
        _os.chdir(cwd)

        # Then
        sorted_csv_nodes = sorted(list(network_csv.nodes))
        sorted_excel_nodes = sorted(list(network_excel.nodes))
        for i in range(len(sorted_csv_nodes)):
            sorted_csv_nodes[i] = str(sorted_csv_nodes[i])
            sorted_excel_nodes[i] = str(sorted_excel_nodes[i])
        self.assertListEqual(sorted_csv_nodes, sorted_excel_nodes)
        assert network_csv.results == network_excel.results
        # signals
        check_dataframe_assertion([], network_csv.timeincrement, network_excel.timeincrement)
        assert (network_csv.timeindex == network_excel.timeindex).all()



import numpy as np
import pandas as pd

import optihood.IO.readers as _rd


def create_scenario_file(configFilePath, numberOfBuildings):
    """
    function to create the input excel file from a config file
    saves the generated excel file at the path given by excel_file_path
    """

    configData = _rd.parse_config(configFilePath)

    excel_data = {}
    buses, columnNames, sheetToSection, updatedLabels = get_excel_variable_names(isGroup=True)

    efficiencyLinks = {'ellink': 0.9999, 'shlink': 0.9, 'dhwlink': 0.9}

    temp_h = {}  # to store temp_h values for stratified storage parameters sheet
    FeedinTariff = 0
    profiles = pd.DataFrame(columns=['name', 'path'])

    for sheet in columnNames:
        excel_data[sheet] = pd.DataFrame(columns=columnNames[sheet])
        if sheet == 'stratified_storage':
            newRow = pd.DataFrame(columns=columnNames[sheet])
            newRow.at[0, 'label'] = list(temp_h)[0]
            newRow.at[0, 'temp_h'] = list(temp_h.values())[0]
            newRow.at[1, 'label'] = list(temp_h)[1]
            newRow.at[1, 'temp_h'] = list(temp_h.values())[1]
        for item in configData[sheetToSection[sheet].lower()]:
            type = item[0]
            if item[1] in ['True', 'False']:  # defines whether or not a component is to be added
                if item[1] == 'True':
                    newRow = pd.DataFrame(columns=columnNames[sheet])
                    newRow.at[0, 'label'] = updatedLabels[item[0]]
                    newRow.at[0, 'active'] = int(bool(item[1] == 'True'))
                    # sheet specific options
                    cols = newRow.columns
                    if 'building' in cols:
                        newRow.at[0, 'building'] = 1
                    if 'to' in cols:
                        newRow.at[0, 'to'] = buses[type][0]
                    if 'from' in cols:
                        newRow.at[0, 'from'] = buses[type][1]
                    if 'connect' in cols:
                        newRow.at[0, 'connect'] = buses[type][2]
                    if type in ['shlink', 'dhwlink']:
                        params = configData['thermallink']
                    else:
                        params = configData[type]
                    if sheet == 'commodity_sources':
                        index = [x for x, y in enumerate(params) if y[0] == 'cost'][0]
                        newRow.at[0, 'variable costs'] = params[index][1]
                        index = [x for x, y in enumerate(params) if y[0] == 'impact'][0]
                        newRow.at[0, 'CO2 impact'] = params[index][1]
                        if type == 'electricityresource':
                            index = [x for x, y in enumerate(params) if y[0] == 'feedintariff'][0]
                            FeedinTariff = -float(params[index][1])
                    else:
                        if sheet == 'links':
                            if not [x for x, y in enumerate(params) if y[0] == 'efficiency']:
                                newRow.at[0, 'efficiency'] = efficiencyLinks[type]
                            newRow.at[0, 'investment'] = 1
                        for p in params:
                            if sheet == 'transformers' and p[0] == 'capacity_max':
                                newRow.at[0, 'capacity_DHW'] = p[1]
                                newRow.at[0, 'capacity_SH'] = p[1]
                                if updatedLabels[item[0]] == 'CHP':
                                    newRow.at[0, 'capacity_el'] = str(round(
                                        float(p[1]) * float(newRow.at[0, 'efficiency'].split(',')[0]) / float(
                                            newRow.at[0, 'efficiency'].split(',')[1]), 0))
                            elif sheet == 'storages' and p[0] == 'temp_h':
                                temp_h[newRow.at[0, 'label']] = p[1]
                            else:
                                newRow.at[0, p[0]] = p[1]
                    excel_data[sheet] = pd.concat([excel_data[sheet], newRow])
            elif 'path' in type:  # weather path or demand path
                profiles = pd.concat([profiles, pd.DataFrame({'name': [updatedLabels[type]], 'path': [item[1]]})])
            elif sheet == 'demand':
                newRow = pd.DataFrame(columns=columnNames[sheet])
                for d in ['electricityDemand', 'spaceHeatingDemand', 'domesticHotWaterDemand']:
                    newRow.at[0, 'label'] = d
                    newRow.at[0, 'from'] = buses[d.lower()][1]
                    if d == 'spaceHeatingDemand' and item[1] == 0:
                        newRow.at[0, 'fixed'] = item[1]
                        newRow.at[0, 'building model'] = 'Yes'
                    else:
                        newRow.at[0, 'fixed'] = 1
                    excel_data[sheet] = pd.concat([excel_data[sheet], newRow])
                excel_data[sheet]['building'] = 1
                excel_data[sheet]['active'] = 1
                excel_data[sheet]['nominal value'] = 1
            elif sheet == 'stratified_storage':
                newRow.at[0, type] = item[1]
                newRow.at[1, type] = item[1]
            else:  # common parameters of all the components of that section
                excel_data[sheet][type] = item[1]
        if sheet == 'stratified_storage':
            excel_data[sheet] = pd.concat([excel_data[sheet], newRow])

    excel_data['profiles'] = profiles

    add_grid_connections_to_excel_data(excel_data, isGroup=True)

    add_busses_to_excel_data(FeedinTariff, excel_data)

    add_buildings_to_excel_data(excel_data, numberOfBuildings)

    return excel_data


def get_excel_variable_names(isGroup=False):
    columnNames = {'commodity_sources': ['label', 'building', 'to', 'variable costs', 'CO2 impact', 'active'],
                   # column names are defined for each sheet (key of dict) in the excel file
                   'solar': ['label', 'building', 'from', 'to', 'connect', 'electrical_consumption',
                             'peripheral_losses', 'latitude', 'longitude', 'tilt', 'azimuth', 'eta_0', 'a_1', 'a_2',
                             'temp_collector_inlet', 'delta_temp_n', 'capacity_max', 'capacity_min', 'lifetime',
                             'maintenance', 'installation', 'planification', 'invest_base', 'invest_cap',
                             'heat_impact', 'elec_impact', 'impact_cap'],
                   'demand': ['label', 'building', 'active', 'from', 'fixed', 'nominal value', 'building model'],
                   'transformers': ['label', 'building', 'active', 'from', 'to', 'efficiency', 'capacity_DHW',
                                    'capacity_SH', 'capacity_el', 'capacity_min', 'lifetime', 'maintenance',
                                    'installation', 'planification', 'invest_base', 'invest_cap', 'heat_impact',
                                    'elec_impact', 'impact_cap'],
                   'storages': ['label', 'building', 'active', 'from', 'to', 'efficiency inflow',
                                'efficiency outflow', 'initial capacity', 'capacity min', 'capacity max',
                                'capacity loss', 'lifetime', 'maintenance', 'installation', 'planification',
                                'invest_base', 'invest_cap', 'heat_impact', 'elec_impact', 'impact_cap'],
                   'stratified_storage': ['label', 'diameter', 'temp_h', 'temp_c', 'temp_env',
                                          'inflow_conversion_factor', 'outflow_conversion_factor', 's_iso',
                                          'lamb_iso', 'alpha_inside', 'alpha_outside']
                   }

    buses = {'naturalgasresource': ['naturalGasBus', '', ''], 'electricityresource': ['gridBus', '', ''],
             'solarcollector': ['dhwStorageBus', 'electricityInBus', 'solarConnectBus'],
             # [to, from, connect] columns in the excel file
             'pv': ['electricityProdBus', '', ''], 'electricitydemand': ['', 'electricityInBus', ''],
             'spaceheatingdemand': ['', 'shDemandBus', ''],
             'domestichotwaterdemand': ['', 'dhwDemandBus', ''],
             'gasboiler': ['shSourceBus,dhwStorageBus', 'naturalGasBus', ''],
             'electricrod': ['shSourceBus,dhwStorageBus', 'electricityInBus', ''],
             'chp': ['electricityProdBus,shSourceBus,dhwStorageBus', 'naturalGasBus', ''],
             'ashp': ['shSourceBus,dhwStorageBus', 'electricityInBus', ''],
             'gshp': ['shSourceBus,dhwStorageBus', 'electricityInBus', ''],
             'electricalstorage': ['electricityBus', 'electricityProdBus', ''],
             'shstorage': ['spaceHeatingBus', 'shSourceBus', ''],
             'dhwstorage': ['domesticHotWaterBus', 'dhwStorageBus', '']}

    sheetToSection = {'commodity_sources': 'CommoditySources', 'solar': 'Solar', 'demand': 'Demands',
                      'transformers': 'Transformers', 'storages': 'Storages', 'stratified_storage': 'StratifiedStorage'}

    updatedLabels = {'weatherpath': 'weather_data', 'path': 'demand_profiles', 'ashp': 'HP', 'gshp': 'GWHP',
                     'electricityresource': 'electricityResource', 'naturalgasresource': 'naturalGasResource',
                     'chp': 'CHP', 'gasboiler': 'GasBoiler', 'electricrod': 'ElectricRod', 'pv': 'pv',
                     'solarcollector': 'solarCollector', 'electricalstorage': 'electricalStorage',
                     'shstorage': 'shStorage', 'dhwstorage': 'dhwStorage', 'stratifiedstorage': 'StratifiedStorage'}

    if isGroup:
        columnNames['links'] = ['label', 'active', 'efficiency', 'invest_base', 'invest_cap', 'investment']
        sheetToSection['links'] = 'Links'
        updatedLabels |= {'ellink': 'electricityLink', 'shlink': 'shLink', 'dhwlink': 'dhwLink'}

    return buses, columnNames, sheetToSection, updatedLabels


def add_grid_connections_to_excel_data(excel_data, isGroup=False):
    label = ['gridElectricity', 'electricitySource', 'producedElectricity', 'shSource', 'spaceHeating']
    building = [1, 1, 1, 1, 1]
    fromNode = ['gridBus', 'electricityProdBus', 'electricityBus', 'shSourceBus', 'spaceHeatingBus']
    toNode = ['electricityInBus', 'electricityBus', 'electricityInBus', 'spaceHeatingBus', 'shDemandBus']
    efficiency = [1, 1, 1, 1, 1]

    if isGroup:
        # does the order matter?
        label.append('domesticHotWater')
        building.append(1)
        fromNode.append('domesticHotWaterBus')
        toNode.append('dhwDemandBus')
        efficiency.append(1)

    excel_data['grid_connection'] = pd.DataFrame(
        {'label': label,
         'building': building,
         'from': fromNode,
         'to': toNode,
         'efficiency': efficiency})


def add_busses_to_excel_data(FeedinTariff, excel_data):
    df = pd.DataFrame(columns=['buses'])
    for sheet, data in excel_data.items():
        for col in ['from', 'to', 'connect']:
            if col in data.columns:
                newBuses = pd.DataFrame(columns=['buses'])
                newBuses['buses'] = data[col].values
                df = pd.concat([df, newBuses])
    df = df.assign(buses=df.buses.str.split(',')).explode('buses')['buses'].unique()
    df = df[df != '']
    excel_data['buses'] = pd.DataFrame(columns=['label', 'building', 'excess', 'excess costs', 'active'])
    for index in range(len(df)):
        excel_data['buses'].at[index, 'label'] = df[index]
        if df[index] == 'electricityBus' and FeedinTariff != 0:
            excel_data['buses'].at[index, 'excess'] = 1
            excel_data['buses'].at[index, 'excess costs'] = FeedinTariff
    excel_data['buses']['building'] = 1
    excel_data['buses']['active'] = 1
    excel_data['buses'].loc[excel_data['buses']['excess'] != 1, 'excess'] = 0


def add_buildings_to_excel_data(excel_data, numberOfBuildings):
    for sheet, data in excel_data.items():
        if 'building' in data.columns:
            buildingNo = [i for i in range(1, numberOfBuildings + 1)] * len(excel_data[sheet].index)
            excel_data[sheet] = pd.DataFrame(np.repeat(data.values, numberOfBuildings, axis=0), columns=data.columns)
            excel_data[sheet]['building'] = buildingNo



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as tkr

from bokeh.plotting import figure, show
from bokeh.layouts import layout, gridplot
from bokeh.models import DatetimeTickFormatter, HoverTool, Legend
from bokeh.palettes import *
from bokeh.io import output_file

from openpyxl import load_workbook
from datetime import datetime
from dateutil.parser import isoparse
import itertools
import pandas as pd
import os
import xlrd as _xlrd


# This file defines different functions for the plotting of the results of the optimization.
# The plots are made at the end of this file, introducing a .xls file previously created during the optimization.
# An important parameter "optMode" needs to be set when running the code (look in the __main__ fragment)


def monthlyBalance(data, bus, new_legends):
    """
    Function for the definition of the monthly summary of a bus
    :param data: dict type, results from the optimization applied to one bus
    :param bus: str type, bus from which the summary is required
    :param new_legends: dict type, new legends to plot on the graph
    :return:
    """
    building = "__" + bus.split("__")[1]
    data_month = data.resample('1m').sum()
    monthShortNames = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

    while len(data_month) != len(monthShortNames):
        data.drop(data.index[-1], inplace=True)
        data_month = data.resample('1m').sum()
    neg_flow = []
    pos_flow = []
    for i in data.columns:
        a = [i.strip("()").split(", ")]
        if "Bus" in a[0][0]:
            neg_flow.append(i)
        else:
            pos_flow.append(i)
    plt.figure()
    mark = []
    for i in neg_flow:
        plt.bar(monthShortNames, -data_month[i], label=new_legends[i.replace(building, "")], bottom=sum(mark))
        mark.append(-data_month[i])
    mark = []
    for i in pos_flow:
        plt.bar(monthShortNames, data_month[i], label=new_legends[i.replace(building, "")], bottom=sum(mark))
        mark.append(data_month[i])

    ax = plt.gca()
    # Shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # Put a legend to the right of the current axis
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.grid(axis='y')
    if "electricity" in bus or "grid" in bus:
        plt.title("Monthly electricity balance for " + building.replace("__", ""))
    elif "spaceHeating" in bus:
        plt.title("Monthly space heating balance for " + building.replace("__", ""))
    else:
        plt.title("Monthly domestic hot water balance for " + building.replace("__", ""))
    plt.show()


def hourlyDailyPlot(data, bus, palette, new_legends):
    """
    Function for the bokeh plot of hourly and daily balance of a bus
    :param data: list of dict type, results from the optimization
    :param bus: list of str type, buses from which the summary is required
    :param palette: palette type form bokeh.palettes (For example, Category10_8), different types can be found on
    https://docs.bokeh.org/en/latest/docs/reference/palettes.html or
    https://docs.bokeh.org/en/latest/_modules/bokeh/palettes.html
    :param new_legends: dict type, new legends to plot on the graph
    For example, Category10_8
    :return:
    """
    p_figs = []
    p_figs_h = []
    p_figs_d = []
    p_plots = []
    a = []
    b = 1
    bus = [x for x in bus if 'electricityProdBus' not in x and 'electricityInBus' not in x]
    for i in range(0, len(data)):
        building = '__' + bus[i].split("__")[1]

        if "electricity" in bus[i]:
            dt = data[i]
            dt.pop(f"(('electricityProdBus{building}', 'electricitySource{building}'), 'flow')")
            data_day = dt.resample('1d').sum()
            p1 = figure(title="Hourly electricity flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p1.add_layout(Legend(), 'right')
            p1.add_tools(HoverTool(tooltips=[('Time', '@x{%d/%m/%Y %H:%M:%S}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            p2 = figure(title="Daily electricity flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p2.add_layout(Legend(), 'right')
            p2.add_tools(HoverTool(tooltips=[('Date', '@x{%d/%m/%Y}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            if len(p_figs_h) > 0:
                p1.x_range = p_figs_h[0].x_range
                p2.x_range = p_figs_d[0].x_range
            colors = itertools.cycle(palette)
            for j, color in zip(dt.columns, colors):
                p1.line(dt.index, dt[j], legend_label=new_legends[j.replace(building, "")], color=color, line_width=1.5)
                p2.line(data_day.index, data_day[j], legend_label=new_legends[j.replace(building, "")], color=color,
                        line_width=1.5)
            p_figs.append([p1, p2])
            p_figs_h.append(p1)
            p_figs_d.append(p2)
            p_plots.append(p1)
            p_plots.append(p2)

        elif "shSource" in bus[i] or "spaceHeatingBus" in bus[i]:
            dt = data[i]
            data_day = dt.resample('1d').sum()
            p3 = figure(title="Hourly space heating flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p3.add_layout(Legend(), 'right')
            p3.add_tools(HoverTool(tooltips=[('Time', '@x{%d/%m/%Y %H:%M:%S}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            p4 = figure(title="Daily space heating flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p4.add_layout(Legend(), 'right')
            p4.add_tools(HoverTool(tooltips=[('Date', '@x{%d/%m/%Y}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            if len(p_figs_h) > 0:
                p3.x_range = p_figs_h[0].x_range
                p4.x_range = p_figs_d[0].x_range
            colors = itertools.cycle(palette)
            for j, color in zip(dt.columns, colors):
                p3.line(dt.index, dt[j], legend_label=new_legends[j.replace(building, "")], color=color, line_width=1.5)
                p4.line(data_day.index, data_day[j], legend_label=new_legends[j.replace(building, "")], color=color,
                        line_width=1.5)
            p_figs.append([p3, p4])
            p_figs_h.append(p3)
            p_figs_d.append(p4)
            p_plots.append(p3)
            p_plots.append(p4)

        else:
            dt = data[i]
            data_day = dt.resample('1d').sum()

            p5 = figure(title="Hourly domestic hot water flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p5.add_layout(Legend(), 'right')
            p5.add_tools(HoverTool(tooltips=[('Time', '@x{%d/%m/%Y %H:%M:%S}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            p6 = figure(title="Daily domestic hot water flows for " + building.replace("__", ""), x_axis_label="Date",
                        y_axis_label="Energy (kWh)", sizing_mode="scale_both")
            p6.add_layout(Legend(), 'right')
            p6.add_tools(HoverTool(tooltips=[('Date', '@x{%d/%m/%Y}'), ('Energy', '@y{0.00}')],
                                   formatters={'@x': 'datetime'},
                                   mode='mouse'))
            if len(p_figs_h) > 0:
                p5.x_range = p_figs_h[0].x_range
                p6.x_range = p_figs_d[0].x_range
            colors = itertools.cycle(palette)
            for j, color in zip(dt.columns, colors):
                p5.line(dt.index, dt[j], legend_label=new_legends[j.replace(building, "")], color=color, line_width=1.5)
                p6.line(data_day.index, data_day[j], legend_label=new_legends[j.replace(building, "")], color=color,
                        line_width=1.5)
            p_figs.append([p5, p6])
            p_figs_h.append(p5)
            p_figs_d.append(p6)
            p_plots.append(p5)
            p_plots.append(p6)

    for p in p_plots:
        p.xaxis[0].formatter = DatetimeTickFormatter(months="%d %b")
        p.legend.click_policy = "hide"
        p.legend.location = "top_left"
        p.legend.label_text_font_size = "8pt"

    return (p_figs_h, p_figs_d)


def toColor(COLORS, obj=None):
    """
    Function from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Assign a deterministic pseudo-random color to argument.
    If COLORS[obj] is set, return that. For strings, this value depends only
    on the string content, so that same strings always yield the same color.
    :param COLORS: dict of components and their assigned color
    :param obj: any hashable object
    :return: a (r, g, b) color tuple if COLORS[obj] is set, otherwise a hexstring
    """
    try:
        color = tuple(rgb / 255.0 for rgb in COLORS[obj])
    except KeyError:
        # random deterministic color
        import hashlib
        color = '#' + hashlib.sha1(obj.encode()).hexdigest()[-6:]
    return color


def groupHbarPlots(ax, group_size, inner_sep=None):
    """
    Function from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Group bars of a horizontal barplot closer together.
    Given an existing horizontal bar plot handle ax, move bars of a given group size (>=2) closer together,
    reducing the distance within the bars of a group, but increasing the distance between different groups.
    By default, bars are placed within a coordinate system 1 unit apart. The space between two bars has
    size 1 - bar_height, which can be specified in matplotlib (and pandas) using the `width` argument.
    :param ax: matplotlib axis
    :param group_size: int type, how many bars to group together
    :param inner_sep: float type, vertical spacing within group (optional). Default: reduce the distance to a half
    :return:
    """
    handles, labels = ax.get_legend_handles_labels()
    bar_height = handles[0][0].get_height()  # assumption: identical for all

    if inner_sep is None:
        inner_sep = 0.5 * (1 - bar_height)

    for column, handle in enumerate(handles):
        for row, patch in enumerate(handle.patches):
            group_number, row_within_group = divmod(row, group_size)

            group_offset = (group_number * group_size
                            + 0.5 * (group_size - 1) * (1 - inner_sep)
                            - 0.5 * (group_size * bar_height))

            patch.set_y(row_within_group * (bar_height + inner_sep)
                        + group_offset)


def deduplicateLegend(handles, labels):
    """
    Function from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Remove double entries from figure legend.
    :param handles: list of legend entry handles
    :param labels: list of legend entry labels
    :return: (handles, labels) tuple of lists with duplicate labels removed
    """
    new_handles = []
    new_labels = []
    for hdl, lbl in zip(handles, labels):
        if not lbl in new_labels:
            new_handles.append(hdl)
            new_labels.append(lbl)
    # also, sort both lists accordingly
    new_labels, new_handles = (list(t) for t in zip(*sorted(zip(new_labels, new_handles))))
    return (new_handles, new_labels)


def resultingDataDiagram(elBus, shBus, dhwBus, costs, env, COLORS, building, newLegends):
    """
    Function inspired from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Function plotting the different results of the optimization. First, costs will be plotted, then the energy produced,
    comparing energy for electricity bus, sh and dhw bus, and finally the retrieved energy from the storages.
    :param elBus: dict type, results from the optimization applied to one bus.
    :param shBus: dict type, results from the optimization applied to one bus.
    :param dhwBus: dict type, results from the optimization applied to one bus.
    :param costs: dict type, resulting costs from the optimization
    :param env: dict type, resulting environmental impacts from the optimization
    :param COLORS: list type, different colors for the different components of the system
    :param building: str type, name of the building
    :return: Four bar plots and the costs, environmental impacts, production and storage dict created
    """
    production = {}
    storage = {}
    alpha = 0
    if optMode == "group":
        list = []

    for flow in elBus.keys():
        if "Storage" in flow and "out" in newLegends[flow.replace("__Building" + str(building), "")]:
            storage[newLegends[flow.replace("__Building" + str(building), "")]] = [0, 0, sum(elBus[flow])]
        elif optMode == "group" and ("electricityLink" in flow):
            if "_in" in newLegends[flow.replace("__Building" + str(building), "")]:
                if newLegends[flow.replace("__Building" + str(building), "")].replace("_in", "") in list:
                    production[newLegends[flow.replace("__Building" + str(building), "")].replace("_in", "")][2] -= sum(
                        elBus[flow])
                else:
                    production[newLegends[flow.replace("__Building" + str(building), "")].replace("_in", "")] = [0, 0,
                                                                                                                 -sum(
                                                                                                                     elBus[
                                                                                                                         flow])]
                    list.append(newLegends[flow.replace("__Building" + str(building), "")].replace("_in", ""))
            elif "_out" in newLegends[flow.replace("__Building" + str(building), "")]:
                if newLegends[flow.replace("__Building" + str(building), "")].replace("_out", "") in list:
                    production[newLegends[flow.replace("__Building" + str(building), "")].replace("_out", "")][
                        2] += sum(elBus[flow])
                else:
                    production[newLegends[flow.replace("__Building" + str(building), "")].replace("_out", "")] = [0, 0,
                                                                                                                  sum(
                                                                                                                      elBus[
                                                                                                                          flow])]
                    list.append(newLegends[flow.replace("__Building" + str(building), "")].replace("_out", ""))
        elif "Storage" not in flow and "Demand" not in flow:
            production[newLegends[flow.replace("__Building" + str(building), "")]] = [0, 0, sum(elBus[flow])]

            # specific for negative flows
            if "HP_SH" in flow or "HP_DHW" in flow:
                alpha += sum(elBus[flow])

    production["HP"] = [0, 0, -alpha]

    for flow in shBus.keys():
        if "Storage" in flow and "out" in newLegends[flow.replace("__Building" + str(building), "")]:
            storage[newLegends[flow.replace("__Building" + str(building), "")]] = [0, sum(shBus[flow]), 0]
        elif "Storage" not in flow and "Demand" not in flow:
            production[newLegends[flow.replace("__Building" + str(building), "")]] = [0, sum(shBus[flow]), 0]

    for flow in dhwBus.keys():
        if "Storage__" in flow and "out" in newLegends[flow.replace("__Building" + str(building), "")]:
            storage[newLegends[flow.replace("__Building" + str(building), "")]] = [sum(dhwBus[flow]), 0, 0]
        elif "Storage__" not in flow and "Demand" not in flow:
            production[newLegends[flow.replace("__Building" + str(building), "")]] = [sum(dhwBus[flow]), 0, 0]

    costs = costs.transpose()
    costs = costs.rename(index={0: building})

    env = env.transpose()
    for i in env.columns:
        for j in newLegends.keys():
            if i.replace("__Building" + str(building), "") in j:
                env = env.rename(columns={i: newLegends[j]})

    production = pd.DataFrame.from_dict(production, orient='index')
    production = production.transpose()

    storage = pd.DataFrame.from_dict(storage, orient='index', columns=["dhw", "sh", "elec"])
    storage = storage.transpose()

    fig = plt.figure(figsize=(18, 8))
    gs = gridspec.GridSpec(1, 4, width_ratios=[3, 4, 8, 3], wspace=0.03)

    ax0 = plt.subplot(gs[0])
    costs_colors = [toColor(COLORS, x) for x in costs.columns]
    bp0 = costs.plot(ax=ax0, kind='barh', color=costs_colors, stacked=True, linewidth=0)

    ax1 = plt.subplot(gs[1])
    env_colors = [toColor(COLORS, x) for x in env.columns]
    bp1 = env.plot(ax=ax1, kind='barh', color=env_colors, stacked=True, linewidth=0)

    ax2 = plt.subplot(gs[2])
    production_colors = [toColor(COLORS, x) for x in production.columns]
    bp2 = production.plot(ax=ax2, kind='barh', color=production_colors, stacked=True, linewidth=0, width=.5)

    ax3 = plt.subplot(gs[3])
    storage_colors = [toColor(COLORS, x) for x in storage.columns]
    bp3 = storage.plot(ax=ax3, kind='barh', color=storage_colors, stacked=True, linewidth=0)

    # remove scenario names from other bar plots
    for ax in [ax1, ax2]:
        ax.set_yticklabels('')
    for ax in [ax2, ax3]:
        groupHbarPlots(ax, group_size=3, inner_sep=0.01)
    ax3.yaxis.tick_right()

    # set limits and ticks for both axes
    for ax in [ax0, ax1, ax2, ax3]:
        ax.yaxis.grid(False)
        ax.xaxis.grid(True, 'major', linestyle='-')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        # group 1,000,000 with commas
        xmin, xmax = ax.get_xlim()
        if xmax > 90 or xmin < -90:
            group_thousands_and_skip_first = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else '{:0,d}'.format(int(x)))
            ax.xaxis.set_major_formatter(group_thousands_and_skip_first)
        else:
            skip_lowest = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else x)
            ax.xaxis.set_major_formatter(skip_lowest)

        # legend
        # set style arguments
        legend_style = {'frameon': False,
                        'loc': 'lower center',
                        'ncol': 2,
                        'bbox_to_anchor': (0.5, .99)}
        # get handels and labels, remove duplicate labels
        handles, labels = deduplicateLegend(*ax.get_legend_handles_labels())
        # set legend to use those
        lg = ax.legend(handles, labels, **legend_style)
        # finally, remove lines from patches
        plt.setp(lg.get_patches(), linewidth=0)

    ax0.set_xlabel('Total costs (CHF)')
    ax1.set_xlabel('Total environmental impacts (kgCo2eq)')
    ax2.set_xlabel('Total energy produced (kWh)')
    ax3.set_xlabel('Retrieved energy (kWh)')
    return fig, costs, env, production, storage


def resultingDataDiagramLoop(elec, sh, dhw, costs, env, colors, buildings, newLegends):
    """
    Function inspired from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Function plotting the graph comparing the different buildings/scenarios on costs, energy produced and energy
    retrieved from storages
    :param elec: list of dict type, optimization results
    :param sh: list of dict type, optimization results
    :param dhw: list of dict type, optimization results
    :param costs: list of dict type, optimization results
    :param env: list of dict type, optimization results
    :param colors: list type, different colors for the different components of the system
    :param buildings: list of str type, name of the different buildings
    :return: figure created
    """
    n_costs, n_env, n_production, n_storage = resultingDataDiagram(elec[0], sh[0], dhw[0], costs[0], env[0], colors,
                                                                   buildings[0], newLegends)[1:]
    for i in range(1, len(elec)):
        a, b, c, d = resultingDataDiagram(elec[i], sh[i], dhw[i], costs[i], env[i], colors, buildings[i], newLegends)[
                     1:]
        n_costs = pd.concat([n_costs, a])
        n_env = pd.concat([n_env, b])
        n_production = pd.concat([n_production, c])
        n_storage = pd.concat([n_storage, d])

    fig = plt.figure(figsize=(18, 8))
    gs = gridspec.GridSpec(1, 4, width_ratios=[4, 3, 8, 3], wspace=0.03)

    ax0 = plt.subplot(gs[0])
    costs_colors = [toColor(colors, x) for x in n_costs.columns]
    bp0 = n_costs.plot(ax=ax0, kind='barh', color=costs_colors, stacked=True, linewidth=0)

    ax1 = plt.subplot(gs[1])
    env_colors = [toColor(colors, x) for x in n_env.columns]
    bp1 = n_env.plot(ax=ax1, kind='barh', color=env_colors, stacked=True, linewidth=0)

    ax2 = plt.subplot(gs[2])
    production_colors = [toColor(colors, x) for x in n_production.columns]
    bp2 = n_production.plot(ax=ax2, kind='barh', color=production_colors, stacked=True, linewidth=0, width=.5)

    ax3 = plt.subplot(gs[3])
    storage_colors = [toColor(colors, x) for x in n_storage.columns]
    bp3 = n_storage.plot(ax=ax3, kind='barh', color=storage_colors, stacked=True, linewidth=0)

    # remove scenario names from other bar plots
    for ax in [ax1, ax2, ax3]:
        ax.set_yticklabels('')
    for ax in [ax2, ax3]:
        groupHbarPlots(ax, group_size=3, inner_sep=0.01)

    # set limits and ticks for both axes
    for ax in [ax0, ax1, ax2, ax3]:
        ax.yaxis.grid(False)
        ax.xaxis.grid(True, 'major', linestyle='-')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        # group 1,000,000 with commas
        xmin, xmax = ax.get_xlim()
        if xmax > 90 or xmin < -90:
            group_thousands_and_skip_first = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else '{:0,d}'.format(int(x)))
            ax.xaxis.set_major_formatter(group_thousands_and_skip_first)
        else:
            skip_lowest = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else x)
            ax.xaxis.set_major_formatter(skip_lowest)

        # legend
        # set style arguments
        legend_style = {'frameon': False,
                        'loc': 'lower center',
                        'ncol': 2,
                        'bbox_to_anchor': (0.5, .99)}
        # get handels and labels, remove duplicate labels
        handles, labels = deduplicateLegend(*ax.get_legend_handles_labels())
        # set legend to use those
        lg = ax.legend(handles, labels, **legend_style)
        # finally, remove lines from patches
        plt.setp(lg.get_patches(), linewidth=0)

    ax0.set_xlabel('Total costs (CHF)')
    ax1.set_xlabel('Total environmental impacts (kgCo2eq)')
    ax2.set_xlabel(
        'Total energy produced (kWh)\n First line: electricity, Second line: space heating,\n Third line: domestic hot water')
    ax3.set_xlabel(
        'Retrieved energy (kWh)\n First line: electricity, Second line: space heating,\n Third line: domestic hot water')

    return fig


def resultingDataDemandDiagram(elBus, shBus, dhwBus, COLORS, building, newLegends):
    """
    Function inspired from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Function plotting the different results of the optimization. First, costs will be plotted, then the energy produced,
    comparing energy for electricity bus, sh and dhw bus, and finally the retrieved energy from the storages.
    :param elBus: dict type, results from the optimization applied to one bus. Called like "solph.views.node(results, bus)"
    :param shBus: dict type, results from the optimization applied to one bus. Called like "solph.views.node(results, bus)"
    :param dhwBus: dict type, results from the optimization applied to one bus. Called like "solph.views.node(results, bus)"
    :param COLORS: list type, different colors for the different components of the system
    :param building: str type, name of the building
    :return: Three bar plots and the elec, sh and dhw dict created
    """
    elec = {}
    sh = {}
    dhw = {}
    if optMode == "group":
        list = []
    for flow in elBus.keys():
        if "Demand" in flow:
            elec[newLegends[flow.replace("__Building" + str(building), "")]] = [0, 0, sum(elBus[flow])]
        elif "producedElectricity" in flow or "gridElectricity" in flow:
            elec[newLegends[flow.replace("__Building" + str(building), "")]] = [0, sum(elBus[flow]), 0]
        elif optMode == "group" and "electricityLink" in flow and "_in" not in newLegends[
            flow.replace("__Building" + str(building), "")]:
            if newLegends[flow.replace("__Building" + str(building), "")] in list:
                elec[newLegends[flow.replace("__Building" + str(building), "")]][0] += sum(elBus[flow])
            else:
                elec[newLegends[flow.replace("__Building" + str(building), "")]] = [sum(elBus[flow]), 0, 0]
                list.append(newLegends[flow.replace("__Building" + str(building), "")])
        else:
            elec[newLegends[flow.replace("__Building" + str(building), "")]] = [sum(elBus[flow]), 0, 0]

        if "Storage" in flow and "in" in newLegends[flow.replace("__Building" + str(building), "")]:
            del elec[newLegends[flow.replace("__Building" + str(building), "")]]
        elif "electricity" not in flow[5:] and "Demand" not in flow:
            del elec[newLegends[flow.replace("__Building" + str(building), "")]]
        elif "Resource" in flow or "excess" in flow:
            del elec[newLegends[flow.replace("__Building" + str(building), "")]]

    for flow in shBus.keys():
        if "Demand" in flow:
            sh[newLegends[flow.replace("__Building" + str(building), "")]] = [0, 0, sum(shBus[flow])]
        else:
            sh[newLegends[flow.replace("__Building" + str(building), "")]] = [0, sum(shBus[flow]), 0]
        if "Storage" in flow and "in" in newLegends[flow.replace("__Building" + str(building), "")]:
            del sh[newLegends[flow.replace("__Building" + str(building), "")]]

    for flow in dhwBus.keys():
        if "Demand" in flow:
            dhw[newLegends[flow.replace("__Building" + str(building), "")]] = [0, 0, sum(dhwBus[flow])]
        else:
            dhw[newLegends[flow.replace("__Building" + str(building), "")]] = [0, sum(dhwBus[flow]), 0]

        if "Storage_" in flow:
            del dhw[newLegends[flow.replace("__Building" + str(building), "")]]

    elec = pd.DataFrame.from_dict(elec, orient='index')
    elec = elec.transpose()

    sh = pd.DataFrame.from_dict(sh, orient='index')
    sh = sh.transpose()

    dhw = pd.DataFrame.from_dict(dhw, orient='index')
    dhw = dhw.transpose()

    fig = plt.figure(figsize=(18, 8))
    gs = gridspec.GridSpec(1, 3, wspace=0.1)

    ax0 = plt.subplot(gs[0])
    elec_colors = [toColor(COLORS, x) for x in elec.columns]
    bp0 = elec.plot(ax=ax0, kind='barh', color=elec_colors, stacked=True, linewidth=0, width=.5)

    ax1 = plt.subplot(gs[1])
    sh_colors = [toColor(COLORS, x) for x in sh.columns]
    bp1 = sh.plot(ax=ax1, kind='barh', color=sh_colors, stacked=True, linewidth=0)

    ax2 = plt.subplot(gs[2])
    dhw_colors = [toColor(COLORS, x) for x in dhw.columns]
    bp2 = dhw.plot(ax=ax2, kind='barh', color=dhw_colors, stacked=True, linewidth=0)

    # remove scenario names from other bar plots
    ax0.set_yticklabels(('', '', 'Building ' + str(building)))
    for ax in [ax1, ax2]:
        ax.set_yticklabels('')
    for ax in [ax0, ax1, ax2]:
        groupHbarPlots(ax, group_size=3, inner_sep=0.0)

    # set limits and ticks for both axes
    for ax in [ax0, ax1, ax2]:
        ax.yaxis.grid(False)
        ax.xaxis.grid(True, 'major', linestyle='-')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        # group 1,000,000 with commas
        xmin, xmax = ax.get_xlim()
        if xmax > 90 or xmin < -90:
            group_thousands_and_skip_first = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else '{:0,d}'.format(int(x)))
            ax.xaxis.set_major_formatter(group_thousands_and_skip_first)
        else:
            skip_lowest = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else x)
            ax.xaxis.set_major_formatter(skip_lowest)

        # legend
        # set style arguments
        legend_style = {'frameon': False,
                        'loc': 'lower center',
                        'ncol': 3,
                        'bbox_to_anchor': (0.5, .99)}
        # get handels and labels, remove duplicate labels
        handles, labels = deduplicateLegend(*ax.get_legend_handles_labels())
        # set legend to use those
        lg = ax.legend(handles, labels, **legend_style)
        # finally, remove lines from patches
        plt.setp(lg.get_patches(), linewidth=0)

    ax0.set_xlabel('Demand and energy produced\n for electricity (kWh)')
    ax1.set_xlabel('Demand and energy produced\n for space heating (kWh)')
    ax2.set_xlabel('Demand and energy produced\n for domestic hot water (kWh)')

    return fig, elec, sh, dhw


def resultingDataDemandDiagramLoop(elec, sh, dhw, colors, buildings, newLegends):
    """
    Function inspired from the URBS platform https://github.com/ojdo/urbs/blob/1house/comp.py
    Function plotting the graph comparing the different buildings/scenarios on costs, energy produced and energy
    retrieved from storages
    :param elec: list of dict type, optimization results
    :param sh: list of dict type, optimization results
    :param dhw: list of dict type, optimization results
    :param colors: list type, different colors for the different components of the system
    :param buildings: list of str type, name of the different buildings
    :return: figure created
    """
    n_elec, n_sh, n_dhw = resultingDataDemandDiagram(elec[0], sh[0], dhw[0], colors, int(buildings[0]))[1:]
    for i in range(1, len(elec)):
        a, b, c = resultingDataDemandDiagram(elec[i], sh[i], dhw[i], colors, buildings[i])[1:]
        n_elec = pd.concat([n_elec, a])
        n_sh = pd.concat([n_sh, b])
        n_dhw = pd.concat([n_dhw, c])

    fig = plt.figure(figsize=(18, 8))
    gs = gridspec.GridSpec(1, 3, wspace=0.1)

    ax0 = plt.subplot(gs[0])
    elec_colors = [toColor(colors, x) for x in n_elec.columns]
    bp0 = n_elec.plot(ax=ax0, kind='barh', color=elec_colors, stacked=True, linewidth=0, width=.5)

    ax1 = plt.subplot(gs[1])
    sh_colors = [toColor(colors, x) for x in n_sh.columns]
    bp1 = n_sh.plot(ax=ax1, kind='barh', color=sh_colors, stacked=True, linewidth=0)

    ax2 = plt.subplot(gs[2])
    dhw_colors = [toColor(colors, x) for x in n_dhw.columns]
    bp2 = n_dhw.plot(ax=ax2, kind='barh', color=dhw_colors, stacked=True, linewidth=0)

    # remove scenario names from other bar plots
    liste = []
    for i in buildings:
        liste.append('')
        liste.append('')
        liste.append('Building ' + str(i))
    ax0.set_yticklabels(liste)
    for ax in [ax1, ax2]:
        ax.set_yticklabels('')
    for ax in [ax0, ax1, ax2]:
        groupHbarPlots(ax, group_size=3, inner_sep=0.0)

    # set limits and ticks for both axes
    for ax in [ax0, ax1, ax2]:
        ax.yaxis.grid(False)
        ax.xaxis.grid(True, 'major', linestyle='-')
        ax.xaxis.set_ticks_position('none')
        ax.yaxis.set_ticks_position('none')

        # group 1,000,000 with commas
        xmin, xmax = ax.get_xlim()
        if xmax > 90 or xmin < -90:
            group_thousands_and_skip_first = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else '{:0,d}'.format(int(x)))
            ax.xaxis.set_major_formatter(group_thousands_and_skip_first)
        else:
            skip_lowest = tkr.FuncFormatter(
                lambda x, pos: '' if pos == 0 else x)
            ax.xaxis.set_major_formatter(skip_lowest)

        # legend
        # set style arguments
        legend_style = {'frameon': False,
                        'loc': 'lower center',
                        'ncol': 3,
                        'bbox_to_anchor': (0.5, .99)}
        # get handels and labels, remove duplicate labels
        handles, labels = deduplicateLegend(*ax.get_legend_handles_labels())
        # set legend to use those
        lg = ax.legend(handles, labels, **legend_style)
        # finally, remove lines from patches
        plt.setp(lg.get_patches(), linewidth=0)

    ax0.set_xlabel('Demand and energy produced\n for electricity (kWh)')
    ax1.set_xlabel('Demand and energy produced\n for space heating (kWh)')
    ax2.set_xlabel('Demand and energy produced\n for domestic hot water (kWh)')

    return fig


def getData(filepath):
    """
    Function for the results recovery from an Excel file
    :param filepath: path to the Excel containing the results of the optimization
    :return: the different dicts created during the optimization
    """
    dict_sheet = {}
    if "xlsx" in filepath:
        wb = load_workbook(filepath, read_only=True, keep_links=False)
        sheet_names = wb.sheetnames
        engine = 'openpyxl'
    elif "xls" in filepath:
        wb = _xlrd.open_workbook(filepath)
        sheet_names = wb.sheet_names()
        engine = 'xlrd'
    else:
        raise NotImplementedError

    for sheet in sheet_names:
        dict_sheet[sheet] = pd.read_excel(filepath, sheet_name=sheet, index_col=0, engine=engine)

    return dict_sheet


def loadPlottingData(resultFilePath, numberOfBuildings):
    """
    Function for loading the data from excel file into variables
    :param resultFilePath: path to the Excel containing the results of the optimization
    :return: the different variables created after reading the excel file
    """
    buses = getData(resultFilePath)
    elec_names = []
    elec_dict = []
    sh_names = []
    sh_dict = []
    dhw_names = []
    dhw_dict = []
    costs_names = []
    costs_dict = []
    env_names = []
    env_dict = []

    buildings_dict = []
    buildings_names = []
    buildings_number = []
    beta = list(buses.keys())
    for i in beta:
        alpha = []
        alpha_a = []
        # Merged buses, if present, are assigned to Building 1 for plotting
        if i in ["electricityBus", "electricityInBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeatingBus",
                 "shDemandBus"]:
            i = i + f'__Building{building}'
        building = i.split("Building")[1]  # building number
        beta_bis = beta.copy()
        firstElBus = True
        firstShBus = True
        firstDhwBus = True
        for j in beta:
            if any('Building' in v for v in beta) and any(f'Building{building}' in v for v in beta):
                j_buses = j
                # Merged buses, if present, are assigned to Building 1 for plotting
                if j in ["electricityBus", "electricityInBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeatingBus",
                         "shDemandBus"]:
                    j = j + f'__Building{building}'
                if j.endswith(building) and (
                        "electricityBus" in j or "electricityInBus" in j or "electricityProdBus" in j or "spaceHeatingBus" in j or "shDemandBus" in j or "shSourceBus" in j or "domesticHotWaterBus" in j or "dhwDemandBus" in j):
                    """if "shDemandBus" not in j and "shSourceBus" not in j and "dhwDemandBus" not in j:
                        alpha.append(j)"""
                    if firstElBus and "electricity" in j:
                        alpha.append(f"electricityBus__Building{building}")
                        firstElBus = False
                    elif firstShBus and ("spaceHeatingBus" in j or "shDemandBus" in j or "shSourceBus" in j):
                        alpha.append(f"spaceHeatingBus__Building{building}")
                        firstShBus = False
                    elif firstDhwBus and ("domesticHotWaterBus" in j or "dhwDemandBus" in j):
                        alpha.append(f"domesticHotWaterBus__Building{building}")
                        firstDhwBus = False
                    if j_buses != j:
                        building_df = buses[j_buses].loc[:, buses[j_buses].columns.str.contains(
                            f'Building{building}')]  # extract columns related to a given building
                        if building == '1':
                            building_df = pd.concat(
                                [building_df, buses[j_buses].loc[:, ~buses[j_buses].columns.str.contains(f'Building')]],
                                axis=1)
                    else:
                        building_df = buses[j_buses]
                    if ("electricityBus" in j) or ("electricityInBus" in j) or ("electricityProdBus" in j):
                        # find the index of electricity bus to merge the results of electricityBus, electricityInBus and electricityProdBus
                        index = -1
                        for ind in range(len(alpha_a)):
                            if "electricityBus" in alpha_a[ind].keys()[0] or "electricityInBus" in alpha_a[ind].keys()[
                                0] or "electricityProdBus" in alpha_a[ind].keys()[0]:
                                index = ind
                        if index != -1:
                            alpha_a[index] = pd.concat((alpha_a[index], building_df), axis=1)
                        else:
                            alpha_a.append(building_df)
                    elif ("spaceHeatingBus" in j) or ("shDemandBus" in j) or ("shSourceBus" in j):
                        # find the index of spaceHeatingBus to merge the results of spaceHeatingBus, shDemandBus and shSourceBus
                        index = -1
                        for ind in range(len(alpha_a)):
                            if "spaceHeatingBus" in alpha_a[ind].keys()[0] or "shDemandBus" in alpha_a[ind].keys()[
                                0] or "shSourceBus" in alpha_a[ind].keys()[0]:
                                index = ind
                        if index != -1:
                            alpha_a[index] = pd.concat((alpha_a[index], building_df), axis=1)
                        else:
                            index = len(alpha_a)
                            alpha_a.append(building_df)
                        flows_to_delete = [
                            f"(('spaceHeating__Building{building}', 'shDemandBus__Building{building}'), 'flow')",
                            f"(('shSourceBus__Building{building}', 'shSource__Building{building}'), 'flow')",
                            f"(('spaceHeating', 'shDemandBus'), 'flow')"]
                        for flow in flows_to_delete:
                            if flow in alpha_a[index].columns:
                                alpha_a[index].pop(flow)
                    else:
                        # find the index of domesticHotWaterBus to merge the results of domesticHotWaterBus and dhwDemandBus
                        index = -1
                        for ind in range(len(alpha_a)):
                            if "domesticHotWaterBus" in alpha_a[ind].keys()[0] or "dhwDemandBus" in alpha_a[ind].keys()[
                                0]:
                                index = ind
                        if index != -1:
                            alpha_a[index] = pd.concat((alpha_a[index], building_df), axis=1)
                        else:
                            index = len(alpha_a)
                            alpha_a.append(building_df)
                        flows_to_delete = [
                            f"(('domesticHotWater__Building{building}', 'dhwDemandBus__Building{building}'), 'flow')",
                            f"(('domesticHotWater', 'dhwDemandBus'), 'flow')"]
                        for flow in flows_to_delete:
                            if flow in alpha_a[index].columns:
                                alpha_a[index].pop(flow)
                    beta_bis.remove(j_buses)
        if alpha != []:
            buildings_dict.append(alpha_a)
            buildings_names.append(alpha)
            buildings_number.append(building)
        """for a in alpha:
            beta.remove(a)"""
        beta = [x for x in beta if f"Building{building}" not in x]

    firstShBus = True
    firstDhwBus = True
    mergedElBuses = []
    mergedShBuses = []
    mergedDhwBuses = []
    mergedLinks = False
    b = 1  # initial building number
    for i in buses.keys():
        i_buses = i
        if 'Building' in i:
            b = int(i.split("Building")[1])  # building number
        # Merged buses, if present, are assigned to Building 1 for plotting
        if i in ["electricityBus", "electricityInBus", "domesticHotWaterBus", "dhwDemandBus", "spaceHeatingBus",
                 "shDemandBus"]:
            i = i + f'__Building{b}'
        tt = {}

        if i_buses != i:
            if 'electricity' in i:
                mergedElBuses.append(i_buses)
            elif 'spaceHeating' in i or 'shDemand' in i:
                mergedShBuses.append(i_buses)
            else:
                mergedDhwBuses.append(i_buses)
            mergedLinks = True
            building_df = buses[i_buses].loc[:, buses[i_buses].columns.str.contains(
                f'Building{b}')]  # extract columns related to a given building
            if b == 1:
                building_df = pd.concat(
                    [building_df, buses[i_buses].loc[:, ~buses[i_buses].columns.str.contains(f'Building')]], axis=1)
        else:
            building_df = buses[i_buses]

        if ("electricityBus" in i) or ("electricityInBus" in i) or ("electricityProdBus" in i):
            if b != 1 and mergedLinks:
                for elBusName in mergedElBuses:
                    elec_names.append(elBusName + f"__Building{b}")
                    mergedbus_df = buses[elBusName].loc[:, buses[elBusName].columns.str.contains(f'Building{b}')]
                    if b <= len(
                            elec_dict):  # to merge the data of the two electricity buses of the same  (elec_dict[0] should have all the electricity related data of building 1, and so on...)
                        elec_dict[b - 1] = pd.concat((elec_dict[b - 1], mergedbus_df), axis=1)
                    else:
                        elec_dict.append(mergedbus_df)
            elec_names.append(i)
            if b <= len(
                    elec_dict):  # to merge the data of the two electricity buses of the same  (elec_dict[0] should have all the electricity related data of building 1, and so on...)
                elec_dict[b - 1] = pd.concat((elec_dict[b - 1], building_df), axis=1)
            else:
                elec_dict.append(building_df)

        if ("spaceHeating" in i) or ("shDemand" in i) or ("shSource" in i):
            if any(str(b) not in v for v in sh_names):
                firstShBus = True
            if b != 1 and mergedLinks:
                for shBusName in mergedShBuses:
                    mergedbus_df = buses[shBusName].loc[:, buses[shBusName].columns.str.contains(f'Building{b}')]
                    if firstShBus:
                        sh_names.append(i)
                        sh_dict.append(mergedbus_df)
                        firstShBus = False
                    else:
                        ind = len(sh_names) - 1
                        sh_dict[ind] = pd.concat((sh_dict[ind], mergedbus_df), axis=1)
                        flows_to_delete = [f"(('spaceHeating__Building{b}', 'shDemandBus__Building{b}'), 'flow')",
                                           f"(('shSourceBus__Building{b}', 'shSource__Building{b}'), 'flow')",
                                           f"(('spaceHeating', 'shDemandBus'), 'flow')"]
                        for flow in flows_to_delete:
                            if flow in sh_dict[ind].columns:
                                sh_dict[ind].pop(flow)
            if firstShBus:
                sh_names.append(i)
                sh_dict.append(building_df)
                firstShBus = False
            else:
                ind = len(sh_names) - 1
                sh_dict[ind] = pd.concat((sh_dict[ind], building_df), axis=1)
                flows_to_delete = [f"(('spaceHeating__Building{b}', 'shDemandBus__Building{b}'), 'flow')",
                                   f"(('shSourceBus__Building{b}', 'shSource__Building{b}'), 'flow')",
                                   f"(('spaceHeating', 'shDemandBus'), 'flow')"]
                for flow in flows_to_delete:
                    if flow in sh_dict[ind].columns:
                        sh_dict[ind].pop(flow)
        if ("domesticHotWater" in i) or ("dhwDemand" in i):
            if any(str(b) not in v for v in dhw_names):
                firstDhwBus = True
            if firstDhwBus:
                dhw_names.append(i)
                dhw_dict.append(building_df)
                firstDhwBus = False
            else:
                ind = len(dhw_names) - 1
                dhw_dict[ind] = pd.concat((dhw_dict[ind], building_df), axis=1)
                flows_to_delete = [f"(('domesticHotWater__Building{b}', 'dhwDemandBus__Building{b}'), 'flow')",
                                   f"(('domesticHotWater', 'dhwDemandBus'), 'flow')"]
                for flow in flows_to_delete:
                    if flow in dhw_dict[ind].columns:
                        dhw_dict[ind].pop(flow)
        if b != 1 and mergedLinks and "dhwStorageBus" in i:
            if any(str(b) not in v for v in dhw_names):
                firstDhwBus = True
            for dhwBusName in mergedDhwBuses:
                mergedbus_df = buses[dhwBusName].loc[:, buses[dhwBusName].columns.str.contains(f'Building{b}')]
                if firstDhwBus:
                    dhw_names.append(f'domesticHotWaterBus__Building{b}')
                    dhw_dict.append(mergedbus_df)
                    firstDhwBus = False
                else:
                    ind = len(dhw_names) - 1
                    dhw_dict[ind] = pd.concat((dhw_dict[ind], mergedbus_df), axis=1)
                    flows_to_delete = [f"(('domesticHotWater__Building{b}', 'dhwDemandBus__Building{b}'), 'flow')",
                                       f"(('domesticHotWater', 'dhwDemandBus'), 'flow')"]
                    for flow in flows_to_delete:
                        if flow in dhw_dict[ind].columns:
                            dhw_dict[ind].pop(flow)
        if "costs" in i:
            costs_names.append(i)
            costs_dict.append(buses[i_buses])
        if "env_impacts" in i:
            env_names.append(i)
            env_dict.append(buses[i_buses])

    return buses, elec_names, elec_dict, sh_names, sh_dict, dhw_names, dhw_dict, costs_names, costs_dict, env_names, env_dict, \
        buildings_dict, buildings_names, buildings_number


def createPlot(resultFilePath, basePath, numberOfBuildings, plotLevel, plotType, flowType, plotAnnualHorizontalBar,
               newLegends):
    # load the plotting data from excel file into variables
    buses, elec_names, elec_dict, sh_names, sh_dict, dhw_names, dhw_dict, costs_names, costs_dict, env_names, env_dict, \
        buildings_dict, buildings_names, buildings_number = loadPlottingData(resultFilePath, numberOfBuildings)

    if flowType == "all":
        if plotType == "energy balance":
            names = elec_names + sh_names + dhw_names
        elif plotType == "bokeh":
            names = buildings_names
            dict = buildings_dict
    elif flowType == "electricity":
        names = elec_names
        if plotType == "bokeh":
            dict = elec_dict
    elif flowType == "space heat":
        names = sh_names
        if plotType == "bokeh":
            dict = sh_dict
    elif flowType == "domestic hot water":
        names = dhw_names
        if plotType == "bokeh":
            dict = dhw_dict
    else:
        raise ValueError("Illegal value for the parameter flow type")

    months = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
    if plotType == "energy balance":
        if plotLevel == "allMonths":
            for i in names:
                if "electricityBus" not in i:
                    monthlyBalance(buses[i], i, newLegends)
    elif plotType == "bokeh":
        ncols = 2
        if plotLevel in months:
            if flowType == "all":
                for i in range(len(buildings_names)):
                    tempDict = dict[i]
                    for j in range(len(tempDict)):
                        tempDict[j] = tempDict[j][tempDict[j].index.strftime('%b') == plotLevel]
                    dict[i] = tempDict
            else:
                for i in range(len(dict)):
                    dict[i] = dict[i][dict[i].index.strftime('%b') == plotLevel]

        elif any(chr.isdigit() for chr in plotLevel):
            if isoparse(plotLevel):
                if flowType == "all":
                    for i in range(len(buildings_names)):
                        tempDict = dict[i]
                        for j in range(len(tempDict)):
                            tempDict[j] = tempDict[j][
                                tempDict[j].index.date == datetime.strptime(plotLevel, "%Y-%m-%d").date()]
                        dict[i] = tempDict
                else:
                    for i in range(len(dict)):
                        dict[i] = dict[i][dict[i].index.date == datetime.strptime(plotLevel, "%Y-%m-%d").date()]

        if flowType == "all":
            plotsHourly = []
            plotsDaily = []
            for i in range(len(buildings_names)):
                plotsH, plotsD = hourlyDailyPlot(dict[i], names[i], Category20_12, newLegends)
                plotsHourly.extend(plotsH)
                plotsDaily.extend(plotsD)
        else:
            plotsHourly, plotsDaily = hourlyDailyPlot(dict, names, Set1_9, newLegends)

        if not os.path.exists(basePath):
            os.makedirs(basePath)

        output_file(os.path.join(basePath, "HourlyBokehPlots.html"))
        grid = gridplot(plotsHourly, ncols=ncols, sizing_mode="scale_both")
        show(grid)
        if not any(chr.isdigit() for chr in plotLevel):
            output_file(os.path.join(basePath, "DailyBokehPlots.html"))
            grid = gridplot(plotsDaily, ncols=ncols, sizing_mode="scale_both")
            show(grid)
    else:
        raise ValueError("Illegal value for the parameter plot type")

    if plotAnnualHorizontalBar == True:
        my_colors = {
            'Battery_out': (100, 160, 200),
            'Battery_in': (100, 160, 200),
            'Storage_sh_out': (196, 7, 27),
            'Storage_dhw_out': (196, 7, 27),
            'HP_dhw': (62, 173, 0),
            'HP_sh': (62, 173, 0),
            'HP': (62, 173, 0),
            'CHP_elec': (0, 101, 189),
            'CHP_sh': (0, 101, 189),
            'CHP_dhw': (128, 0, 128),
            'CHP': (0, 101, 189),
            'Gas_sh': (0, 110, 120),
            'GasBoiler': (0, 110, 120),
            'PV_elec': (0, 128, 90),
            'SolarCollector': (188, 128, 90),
            'Inputs': (252, 93, 93),
            'Operation': (252, 93, 93),
            'Investment': (0, 119, 138),
            'Feed-in': (0, 215, 203),
            'Demand_elec': (131, 166, 151),
            'Demand_sh': (131, 166, 151),
            'Demand_dhw': (131, 166, 151),
            'Grid_purchase': (237, 127, 16)
        }
        if optMode == "group":
            my_colors['electricityLink_in'] = (255, 215, 0)
            my_colors['electricityLink_out'] = (255, 215, 0)
            my_colors['electricityLink'] = (255, 215, 0)

        COLORS = {}
        for name, color in my_colors.items():
            COLORS[name] = color
        """
        for i in range(len(buildings_names)):
            fig1 = resultingDataDiagram(elec_dict[i], sh_dict[i], dhw_dict[i], costs_dict[i], env_dict[i], COLORS, buildings_number[i])[0]
            fig2 = resultingDataDemandDiagram(elec_dict[i], sh_dict[i], dhw_dict[i], COLORS, buildings_number[i])[0]
        """
        fig3 = resultingDataDiagramLoop(elec_dict, sh_dict, dhw_dict, costs_dict, env_dict, COLORS, buildings_number,
                                        newLegends)
        fig4 = resultingDataDemandDiagramLoop(elec_dict, sh_dict, dhw_dict, COLORS, buildings_number, newLegends)
        plt.show()


def plot(excelFileName, figureFilePath, numberOfBuildings, plotLevel, plotType, flowType, plotlabels='default',
         plotAnnualHorizontalBar=False):
    #####################################
    ########## Classic plots  ###########
    #####################################
    # New legends need to be defined by hand for each new flow added to the energy network
    if plotlabels == 'default':
        newLegends = {
            "(('naturalGasResource', 'naturalGasBus'), 'flow')": "NaturalGas",
            "(('qSource', 'heatSourceBus'), 'flow')": "qSource",
            "(('electricityBus', 'excesselectricityBus'), 'flow')": "Feed-in",
            "(('electricityProdBus', 'excesselectricityProdBus'), 'flow')": "Feed-in",
            "(('electricityProdBus', 'electricalStorage'), 'flow')": "Battery_in",
            "(('electricityInBus', 'electricityDemand'), 'flow')": "Demand_elec",
            "(('electricityInBus', 'emobilityDemand'), 'flow')": "Demand_mobility",
            "(('electricityInBus', 'HP'), 'flow')": "HP",
            "(('electricityInBus', 'GWHP'), 'flow')": "GWHP",
            "(('CHP', 'electricityProdBus'), 'flow')": "CHP_elec",
            "(('electricalStorage', 'electricityBus'), 'flow')": "Battery_out",
            "(('electricitySource', 'electricityBus'), 'flow')": "Elect_produced (not stored)",
            "(('electricityResource', 'gridBus'), 'flow')": "Grid_purchase",
            "(('dhwStorage', 'domesticHotWaterBus'), 'flow')": "Storage_dhw_out",
            "(('dhwStorageBus', 'dhwStorage'), 'flow')": "Storage_dhw_in",
            "(('domesticHotWaterBus', 'domesticHotWater'), 'flow')": "dhw_direct_to_load",
            "(('dhwDemandBus', 'domesticHotWaterDemand'), 'flow')": "Demand_dhw",
            "(('HP', 'dhwStorageBus'), 'flow')": "HP_dhw",
            "(('GWHP', 'dhwStorageBus'), 'flow')": "GWHP_dhw",
            "(('CHP', 'dhwStorageBus'), 'flow')": "CHP_dhw",
            "(('shSource', 'spaceHeatingBus'), 'flow')": "SH_produced (not stored)",
            "(('shStorage', 'spaceHeatingBus'), 'flow')": "Storage_sh_out",
            "storage_content": "Storage_content",
            "(('shSourceBus', 'shStorage'), 'flow')": "Storage_sh_in",
            "(('spaceHeatingBus', 'spaceHeating'), 'flow')": "SH_direct_to_load",
            "(('shDemandBus', 'spaceHeatingDemand'), 'flow')": "Demand_sh",
            "(('shDemandBus', 'excessshDemandBus'), 'flow')": "Excess SH production",
            "(('shSourceBus', 'excessshSourceBus'), 'flow')": "Excess SH production",
            "(('CHP', 'shSourceBus'), 'flow')": "CHP_sh",
            "(('GasBoiler', 'shSourceBus'), 'flow')": "Gas_sh",
            "(('GasBoiler', 'dhwStorageBus'), 'flow')": "Gas_dhw",
            "(('HP', 'shSourceBus'), 'flow')": "HP_sh",
            "(('GWHP', 'shSourceBus'), 'flow')": "GWHP_sh",
            "(('electricityBus', 'producedElectricity'), 'flow')": "Self_consumption",
            "(('gridBus', 'gridElectricity'), 'flow')": "Electricity_grid",
            "(('pv', 'electricityProdBus'), 'flow')": "PV_elec",
            "(('solarCollectorSH', 'dhwStorageBus'), 'flow')": "SolarCollector",
            "(('electricityInBus', 'solarCollectorSH'), 'flow')": "SolarCollector",
            "(('gridElectricity', 'electricityInBus'), 'flow')": "Electricity_grid",
            "(('producedElectricity', 'electricityInBus'), 'flow')": "Self_consumption",
            "(('electricityProdBus', 'electricitySource'), 'flow')": "Battery_bypass",
            "(('domesticHotWaterBus', 'domesticHotWaterDemand'), 'flow')": "Demand_dhw",
            "(('electricityInBus', 'ElectricRod'), 'flow')": "Electric_rod",
            "(('ElectricRod', 'shSourceBus'), 'flow')": "Electric_rod_sh",
            "(('ElectricRod', 'dhwStorageBus'), 'flow')": "Electric_rod_dhw",
            "(('electricityInBus', 'GWHP35'), 'flow')": "GWHP35",
            "(('electricityInBus', 'GWHP60'), 'flow')": "GWHP60",
            "(('GWHP60', 'dhwStorageBus'), 'flow')": "GWHP60_dhw",
            "(('GWHP35', 'shSourceBus'), 'flow')": "GWHP35_sh",
        }
        newLegends["(('electricityBus', 'electricityLink'), 'flow')"] = "electricityLink_out"
        newLegends["(('electricityLink', 'electricityInBus'), 'flow')"] = "electricityLink_in"
        newLegends["(('spaceHeatingBus', 'shLink'), 'flow')"] = "shLink_out"
        newLegends["(('shLink', 'shDemandBus'), 'flow')"] = "shLink_in"
        newLegends["(('domesticHotWaterBus', 'dhwLink'), 'flow')"] = "dhwLink_out"
        newLegends["(('dhwLink', 'dhwDemandBus'), 'flow')"] = "dhwLink_in"
    else:
        newLegends = {
            "(('naturalGasResource', 'naturalGasBus'), 'flow')": plotlabels["naturalGas"],
            "(('qSource', 'heatSourceBus'), 'flow')": plotlabels["qSource"],
            "(('electricityBus', 'excesselectricityBus'), 'flow')": plotlabels["excessEl"],
            "(('electricityProdBus', 'electricalStorage'), 'flow')": plotlabels["StorageEl"] + "_in",
            "(('electricityInBus', 'electricityDemand'), 'flow')": plotlabels["DemandEl"],
            "(('electricityInBus', 'emobilityDemand'), 'flow')": plotlabels["DemandMob"],
            "(('electricityInBus', 'HP'), 'flow')": plotlabels["HP"],
            "(('electricityInBus', 'GWHP'), 'flow')": plotlabels["GWHP"],
            "(('CHP', 'electricityProdBus'), 'flow')": plotlabels["CHP"] + "_el",
            "(('electricalStorage', 'electricityBus'), 'flow')": plotlabels["StorageEl"] + "_out",
            "(('electricitySource', 'electricityBus'), 'flow')": plotlabels["localEl"],
            "(('electricityResource', 'gridBus'), 'flow')": plotlabels["grid"],
            "(('dhwStorage', 'domesticHotWaterBus'), 'flow')": plotlabels["StorageDhw"] + "_out",
            "(('dhwStorageBus', 'dhwStorage'), 'flow')": plotlabels["StorageDhw"] + "_in",
            "(('domesticHotWaterBus', 'domesticHotWater'), 'flow')": "DHW_direct_to_load",
            "(('dhwDemandBus', 'domesticHotWaterDemand'), 'flow')": plotlabels["DemandDhw"],
            "(('HP', 'dhwStorageBus'), 'flow')": plotlabels["HP"] + "_dhw",
            "(('GWHP', 'dhwStorageBus'), 'flow')": plotlabels["GWHP"] + "_dhw",
            "(('CHP', 'dhwStorageBus'), 'flow')": plotlabels["CHP"] + "_dhw",
            "(('shSource', 'spaceHeatingBus'), 'flow')": plotlabels["prodSH"] + " (not stored)",
            "(('shStorage', 'spaceHeatingBus'), 'flow')": plotlabels["StorageSh"] + "_out",
            "storage_content": "Storage_content",
            "(('shSourceBus', 'shStorage'), 'flow')": plotlabels["StorageSh"] + "_in",
            "(('spaceHeatingBus', 'spaceHeating'), 'flow')": "SH_direct_to_load",
            "(('shDemandBus', 'spaceHeatingDemand'), 'flow')": plotlabels["DemandSh"],
            "(('shDemandBus', 'excessshDemandBus'), 'flow')": plotlabels["excessSh"],
            "(('shSourceBus', 'excessshSourceBus'), 'flow')": plotlabels["excessSh"],
            "(('CHP', 'shSourceBus'), 'flow')": plotlabels["CHP"] + "_sh",
            "(('GasBoiler', 'shSourceBus'), 'flow')": plotlabels["gasBoiler"] + "_sh",
            "(('GasBoiler', 'dhwStorageBus'), 'flow')": plotlabels["gasBoiler"] + "_dhw",
            "(('HP', 'shSourceBus'), 'flow')": plotlabels["HP"] + "_sh",
            "(('GWHP', 'shSourceBus'), 'flow')": plotlabels["GWHP"] + "_sh",
            "(('electricityBus', 'producedElectricity'), 'flow')": "Self_consumption",
            "(('gridBus', 'gridElectricity'), 'flow')": plotlabels["grid"],
            "(('pv', 'electricityProdBus'), 'flow')": plotlabels["pv"] + "_el",
            "(('solarCollectorSH', 'dhwStorageBus'), 'flow')": plotlabels["solarCollector"],
            "(('electricityInBus', 'solarCollectorSH'), 'flow')": plotlabels["solarCollector"],
            "(('gridElectricity', 'electricityInBus'), 'flow')": plotlabels["grid"],
            "(('producedElectricity', 'electricityInBus'), 'flow')": "Self_consumption",
            "(('electricityProdBus', 'electricitySource'), 'flow')": plotlabels["StorageEl"] + "_bypass",
            "(('domesticHotWaterBus', 'domesticHotWaterDemand'), 'flow')": plotlabels["DemandDhw"],
            "(('electricityInBus', 'ElectricRod'), 'flow')": plotlabels["ElectricRod"],
            "(('ElectricRod', 'shSourceBus'), 'flow')": plotlabels["ElectricRod"] + "_sh",
            "(('ElectricRod', 'dhwStorageBus'), 'flow')": plotlabels["ElectricRod"] + "_dhw",
            "(('electricityInBus', 'GWHP35'), 'flow')": plotlabels["GWHP"] + "35",
            "(('electricityInBus', 'GWHP60'), 'flow')": plotlabels["GWHP"] + "60",
            "(('GWHP60', 'dhwStorageBus'), 'flow')": plotlabels["GWHP"] + "60_dhw",
            "(('GWHP35', 'shSourceBus'), 'flow')": plotlabels["GWHP"] + "35_sh",
        }
        newLegends["(('electricityBus', 'electricityLink'), 'flow')"] = plotlabels["elBus"] + " Link (out)"
        newLegends["(('electricityLink', 'electricityInBus'), 'flow')"] = plotlabels["elBus"] + " Link (in)"
        newLegends["(('spaceHeatingBus', 'shLink'), 'flow')"] = plotlabels["shBus"] + " Link (out)"
        newLegends["(('shLink', 'shDemandBus'), 'flow')"] = plotlabels["shBus"] + " Link (in)"
        newLegends["(('domesticHotWaterBus', 'dhwLink'), 'flow')"] = plotlabels["dhwBus"] + " Link (out)"
        newLegends["(('dhwLink', 'dhwDemandBus'), 'flow')"] = plotlabels["dhwBus"] + " Link (in)"

    createPlot(excelFileName, figureFilePath, numberOfBuildings, plotLevel, plotType, flowType, plotAnnualHorizontalBar,
               newLegends)


if __name__ == '__main__':
    optMode = "group"  # parameter defining whether the results file corresponds to "indiv" or "group" optimization
    numberOfBuildings = 4
    plotOptim = 1  # defines the number of the optimization to plot
    plotLevel = "allMonths"  # permissible values (for energy balance plot): "allMonths" {for all months}
    # or specific month {"Jan", "Feb", "Mar", etc. three letter abbreviation of the month name}
    # or specific date {format: YYYY-MM-DD}
    plotType = "energy balance"  # permissible values: "energy balance", "bokeh"
    flowType = "electricity"  # permissible values: "all", "electricity", "space heat", "domestic hot water"
    plotAnnualHorizontalBar = False  # determines whether the annual horizontal bar is plot or not

    plot(os.path.join(r"..\data\Results", f"results{numberOfBuildings}_{plotOptim}_{optMode}.xlsx"), r"..\data\Figures",
         plotLevel, plotType, flowType, plotAnnualHorizontalBar)



"""
under-development component for a linear RC model for heating a Building
"""
from optihood._helpers import *
import oemof.solph as solph
from oemof.solph._plumbing import sequence
from pyomo.core.base.block import ScalarBlock
from pyomo.environ import BuildAction
from pyomo.environ import Constraint
from pyomo.environ import Expression
from pyomo.environ import NonNegativeReals, Reals, Binary
from pyomo.environ import Set
from pyomo.environ import Var
import numpy as np

class SinkRCModel(solph.components.Sink):
    """
    Building RC Model with a possibility of heat outflow (through chiller) as well

    Parameters
    ----------
    rDistribution : Thermal resistance between indoor and distribution system states [K/kW]
    cDistribution : Thermal capacity of distribution system state [kWh/K]
    rIndoor : Thermal resistance between indoor and wall states [K/kW]
    cIndoor : Thermal capacity of indoor air state [kWh/K]
    rWall : Thermal resistance between wall state and outside [K/kW]
    cWall  : Thermal capacity of wall state [kWh/K]
    gAreaWindows : g-value multiplied with aperture area of windows [m^2]
    qDistributionMin : Minimum operating power from the SC tank to the distribution system [kW]
    qDistributionMax : Maximum operating power from the SC tank to the distribution system [kW]
    tIndoorMin : Indoor minimum comfort temperature [ºC]
    tIndoorMax : Indoor maximum comfort temperature [ºC]
    tIndoorInit : Indoor initial temperature [ºC]
    tWallInit : Wall initial temperature [ºC]
    tDistributionInit : Distribution system initial temperature [ºC]
    tAmbient : Ambient outside air temperature at each timestep [ºC]
    totalIrradiationHorizontal : Total horizontal irradiation at each timestep [kW/m^2]
    heatGainOccupants : Internal heat gains from occupants at each timestep [kW]
    """

    def __init__(
            self,
            tAmbient,
            totalIrradiationHorizontal,
            heatGainOccupants,
            rDistribution=0.75,
            cDistribution=0.26,
            rIndoor=0.09,
            cIndoor=0.97,
            rWall=1.70,
            cWall=226.30,
            gAreaWindows=1.5,
            qDistributionMin=0,
            qDistributionMax=1000,
            tIndoorMin=19,
            tIndoorMax=23,
            tIndoorInit=21,
            tWallInit=21,
            tDistributionInit=21,
            outputs=None,
            **kwargs,
    ):
        super().__init__(custom_attributes={"outputs": outputs},**kwargs)
        self.rDistribution = rDistribution
        self.cDistribution = cDistribution
        self.rIndoor = rIndoor
        self.cIndoor = cIndoor
        self.rWall = rWall
        self.cWall = cWall
        self.gAreaWindows = gAreaWindows
        self.qDistributionMin = qDistributionMin
        self.qDistributionMax = qDistributionMax
        self.tIndoorMin = tIndoorMin
        self.tIndoorMax = tIndoorMax
        self.tIndoorInit = tIndoorInit
        self.tWallInit = tWallInit
        self.tDistributionInit = tDistributionInit
        self.tAmbient = sequence(tAmbient)
        self.totalIrradiationHorizontal = sequence(totalIrradiationHorizontal)
        self.heatGainOccupants = sequence(heatGainOccupants)

    def constraint_group(self):
        return SinkRCModelBlock

class SinkRCModelBlock(ScalarBlock):
    """
    Constraints for SinkRCModel Class
    """
    CONSTRAINT_GROUP = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _create(self, group=None):
        if group is None:
            return None

        m = self.parent_block()

        # for all Sink RC model components get inflow and outflow
        for n in group:
            n.inflow = list(n.inputs)[0]
            n.outflow = list(n.outputs)[0]

        #  ************* SET OF CUSTOM SINK COMPONENTS *****************************

        # Set of Sink RC model Components
        self.sinkrc = Set(initialize=[n for n in group])

        #  ************* DECISION VARIABLES *****************************

        # Variable indoor temperature
        self.tIndoor = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))
        self.tIndoor_prev = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))
        # Variable wall temperature
        self.tWall = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))
        self.tWall_prev = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))

        # Variable distribution temperature
        self.tDistribution = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))
        self.tDistribution_prev = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))

        # Variable indoor comfort temperature range violation
        self.epsilonIndoor = Var(self.sinkrc, m.TIMESTEPS, within=NonNegativeReals, bounds=(0, 1000))

        #  ************* CONSTRAINTS *****************************

        def _initial_indoor_temperature_rule(block):
            """set initial values of indoor temperature
            """
            for g in group:
                lhs = self.tIndoor_prev[g, 0]
                rhs = g.tIndoorInit
                block.initial_indoor_temperature.add((g, 0), (lhs == rhs))

        self.initial_indoor_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.initial_indoor_temperature_build = BuildAction(rule=_initial_indoor_temperature_rule)

        def _initial_wall_temperature_rule(block):
            """set initial values of wall temperature
            """
            for g in group:
                lhs = self.tWall_prev[g, 0]
                rhs = g.tWallInit
                block.initial_wall_temperature.add((g, 0), (lhs == rhs))

        self.initial_wall_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.initial_wall_temperature_build = BuildAction(rule=_initial_wall_temperature_rule)

        def _initial_distribution_temperature_rule(block):
            """set initial values of distribution temperature
            """
            for g in group:
                lhs = self.tDistribution_prev[g, 0]
                rhs = g.tDistributionInit
                block.initial_distribution_temperature.add((g, 0), (lhs == rhs))

        self.initial_distribution_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.initial_distribution_temperature_build = BuildAction(rule=_initial_distribution_temperature_rule)

        def _prev_indoor_temperature_rule(block):
            """set initial values of indoor temperature
            """
            for g in group:
                for t in m.TIMESTEPS:
                    if t != 0:
                        lhs = self.tIndoor_prev[g, t]
                        rhs = self.tIndoor[g, t-1]
                        block.prev_indoor_temperature.add((g, t), (lhs == rhs))

        self.prev_indoor_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.prev_indoor_temperature_build = BuildAction(rule=_prev_indoor_temperature_rule)

        def _prev_wall_temperature_rule(block):
            """set initial values of wall temperature
            """
            for g in group:
                for t in m.TIMESTEPS:
                    if t != 0:
                        lhs = self.tWall_prev[g, t]
                        rhs = self.tWall[g, t-1]
                        block.prev_wall_temperature.add((g, t), (lhs == rhs))

        self.prev_wall_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.prev_wall_temperature_build = BuildAction(rule=_prev_wall_temperature_rule)

        def _prev_distribution_temperature_rule(block):
            """set initial values of distribution temperature
            """
            for g in group:
                for t in m.TIMESTEPS:
                    if t != 0:
                        lhs = self.tDistribution_prev[g, t]
                        rhs = self.tDistribution[g, t-1]
                        block.prev_distribution_temperature.add((g, t), (lhs == rhs))

        self.prev_distribution_temperature = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.prev_distribution_temperature_build = BuildAction(rule=_prev_distribution_temperature_rule)

        def _indoor_comfort_upper_limit_rule(block):
            """Indoor comfort temperature < = maximum limit
            """
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = self.tIndoor[g, t]
                    rhs = g.tIndoorMax + self.epsilonIndoor[g, t]
                    block.indoor_comfort_upper_limit.add((g, t), (lhs <= rhs))

        self.indoor_comfort_upper_limit = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.indoor_comfort_upper_limit_build = BuildAction(rule=_indoor_comfort_upper_limit_rule)

        def _indoor_comfort_lower_limit_rule(block):
            """Indoor comfort temperature > = minimum limit
            """
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = self.tIndoor[g, t]
                    rhs = g.tIndoorMin[t] - self.epsilonIndoor[g, t]
                    block.indoor_comfort_lower_limit.add((g, t), (lhs >= rhs))

        self.indoor_comfort_lower_limit = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.indoor_comfort_lower_limit_build = BuildAction(rule=_indoor_comfort_lower_limit_rule)

        def _q_distribution_upper_limit_rule(block):
            """q distribution < = maximum limit
            """
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g.inflow, g, t]
                    rhs = g.qDistributionMax
                    block.q_distribution_upper_limit.add((g, t), (lhs <= rhs))

        self.q_distribution_upper_limit = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.q_distribution_upper_limit_build = BuildAction(rule=_q_distribution_upper_limit_rule)

        def _q_distribution_lower_limit_rule(block):
            """q distribution > = minimum limit
            """
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g.inflow, g, t]
                    rhs = g.qDistributionMin
                    block.q_distribution_lower_limit.add((g, t), (lhs >= rhs))

        self.q_distribution_lower_limit = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.q_distribution_lower_limit_build = BuildAction(rule=_q_distribution_lower_limit_rule)

        def _q_out_lower_limit_rule(block):
            """q out > = minimum limit
            """
            for t in m.TIMESTEPS:
                for g in group:
                    lhs = m.flow[g, g.outflow, t]
                    rhs = g.qDistributionMin
                    # rhs = g.qDistributionMin * self.Xsc_dis[g, t]
                    block.q_out_lower_limit.add((g, t), (lhs >= rhs))

        self.q_out_lower_limit = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.q_out_lower_limit_build = BuildAction(rule=_q_out_lower_limit_rule)

        def _indoor_temperature_equation_rule(block):
            """discrete state space equation for tIndoor
            """
            for g in group:
                c2 = 1/(g.rIndoor*g.cIndoor)
                c3 = 1/(g.rDistribution*g.cIndoor)
                c1 = 1 - c2 - c3
                c4 = g.gAreaWindows/g.cIndoor
                c5 = 1/g.cIndoor
                for t in m.TIMESTEPS:
                    lhs = self.tIndoor[g, t]
                    rhs = c1*self.tIndoor_prev[g, t] + c2*self.tWall_prev[g, t] + c3*self.tDistribution_prev[g, t] + c4*g.totalIrradiationHorizontal[t] + c5*g.heatGainOccupants[t]
                    block.indoor_temperature_equation.add((g, t), (lhs == rhs))

        self.indoor_temperature_equation = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.indoor_temperature_equation_build = BuildAction(rule=_indoor_temperature_equation_rule)

        def _wall_temperature_equation_rule(block):
            """discrete state space equation for tWall
            """
            for g in group:
                c1 = 1 / (g.rIndoor * g.cWall)
                c3 = 1 / (g.rWall * g.cWall)
                c2 = 1 - c1 - c3
                for t in m.TIMESTEPS:
                    lhs = self.tWall[g, t]
                    rhs = c1 * self.tIndoor_prev[g, t] + c2 * self.tWall_prev[g, t] + c3 * g.tAmbient[t]
                    block.wall_temperature_equation.add((g, t), (lhs == rhs))

        self.wall_temperature_equation = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.wall_temperature_equation_build = BuildAction(rule=_wall_temperature_equation_rule)

        def _distribution_temperature_equation_rule(block):
            """discrete state space equation for tDistribution
            """
            for g in group:
                c1 = 1 / (g.rDistribution * g.cDistribution)
                c2 = 1 - c1
                c3 = 1 /g.cDistribution
                for t in m.TIMESTEPS:
                    lhs = self.tDistribution[g, t]
                    rhs = c1 * self.tIndoor_prev[g, t] + c2 * self.tDistribution_prev[g, t] + c3 * (m.flow[g.inflow, g, t] - m.flow[g, g.outflow, t])
                    block.distribution_temperature_equation.add((g, t), (lhs == rhs))

        self.distribution_temperature_equation = Constraint(group, m.TIMESTEPS, noruleinit=True)
        self.distribution_temperature_equation_build = BuildAction(rule=_distribution_temperature_equation_rule)

    def _objective_expression(self):
        r"""Objective expression to add costs to high values of self.epsilon"""
        m = self.parent_block()
        variable_costs = 0
        fixed_costs = 0

        for g in self.sinkrc:
            for t in m.TIMESTEPS:
                variable_costs += self.epsilonIndoor[g,t]*2

        self.cost = Expression(expr=variable_costs + fixed_costs)
        return self.cost